"""
Core implementation of :mod:`pytools.fit`.
"""
import logging
from abc import ABCMeta, abstractmethod
from typing import Generic, TypeVar

from ..api import AllTracker

log = logging.getLogger(__name__)

#
# Exported names
#

__all__ = ["NotFittedError", "FittableMixin"]


#
# Type variables
#

T_Self = TypeVar("T_Self")
T_Data = TypeVar("T_Data")


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


class NotFittedError(Exception):
    """
    Raised when a fittable object was expected to be fitted but was not fitted.
    """


class FittableMixin(Generic[T_Data], metaclass=ABCMeta):
    """
    Mix-in class that supports fitting the object to data.
    """

    @abstractmethod
    def fit(self: T_Self, _x: T_Data, **fit_params) -> T_Self:
        """
        Fit this object to the given data.

        :param _x: the data to fit this object to
        :param fit_params: optional fitting parameters
        :return: self
        """
        pass

    @property
    @abstractmethod
    def is_fitted(self) -> bool:
        """
        ``True`` if this object is fitted, ``False`` otherwise.
        """
        pass

    def _ensure_fitted(self) -> None:
        """
        Raise a :class:`.NotFittedError` if this object is not fitted.

        :meta public:
        :raise NotFittedError: this object is not fitted
        """
        if not self.is_fitted:
            raise NotFittedError(f"{type(self).__name__} is not fitted")


__tracker.validate()



"""
Tests for package pytools.viz.dendrogram
"""

# noinspection PyPackageRequirements
import hashlib
import logging
from io import StringIO

import numpy as np

# noinspection PyPackageRequirements
import pytest

# noinspection PyPackageRequirements
import scipy.cluster.hierarchy as hc

from pytools.viz.dendrogram import DendrogramDrawer, DendrogramReportStyle, LinkageTree

log = logging.getLogger(__name__)


@pytest.fixture
def linkage_matrix() -> np.ndarray:
    """Create a linkage matrix."""
    x = np.array([[i] for i in [2, 8, 0, 4, 1, 9, 9, 0]])
    return hc.linkage(x)


@pytest.fixture
def linkage_tree(linkage_matrix: np.ndarray) -> LinkageTree:
    """Create a linkage tree for drawing tests."""
    return LinkageTree(
        scipy_linkage_matrix=linkage_matrix,
        leaf_names=list("ABCDEFGH"),
        leaf_weights=[(w + 1) / 36 for w in range(8)],
    )


def test_dendrogram_drawer_text(linkage_matrix: np.ndarray) -> None:
    checksum_dendrogram_report = "2d94fe5966d1fb77b4216c16e9845da6"
    leaf_names = list("ABCDEFGH")
    leaf_weights = [(w + 1) / 36 for w in range(8)]

    with pytest.raises(ValueError) as value_error:
        LinkageTree(
            scipy_linkage_matrix=linkage_matrix,
            leaf_names=leaf_names,
            leaf_weights=leaf_weights,
            max_distance=1,
        )
    assert value_error.value.args == (
        "arg max_distance=1 must be equal to or greater than the maximum distance "
        "(= 4.0) in the linkage tree",
    )

    linkage_tree = LinkageTree(
        scipy_linkage_matrix=linkage_matrix,
        leaf_names=leaf_names,
        leaf_weights=[(w + 1) / 36 for w in range(8)],
        distance_label="distance",
        leaf_label="label",
        weight_label="weight",
    )

    with StringIO() as out:
        dd = DendrogramDrawer(style=DendrogramReportStyle(out=out))
        dd.draw(data=linkage_tree, title="Test")
        report_str = str(out.getvalue())
        log.debug(f"\n{report_str}")
        assert (
            hashlib.md5(str(report_str).encode("utf-8")).hexdigest()
        ) == checksum_dendrogram_report



"""
Base classes for distribution styles.
"""

import logging
from abc import ABCMeta, abstractmethod
from typing import Sequence

from pytools.api import AllTracker
from pytools.viz import DrawingStyle

log = logging.getLogger(__name__)

#
# Exported names
#

__all__ = ["XYSeries", "ECDF", "ECDFStyle"]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


class XYSeries:
    """
    Series of `x` and `y` coordinates for plotting; `x` and `y` values are stored in two
    separate sequences of the same length.
    """

    def __init__(self, x: Sequence[float], y: Sequence[float]):
        """
        :param x: series of all `x` coordinate values
        :param y: series of all `y` coordinate values
        """
        assert len(x) == len(y), "x and y have the same length"
        self.x = x
        self.y = y

    #: series of all `x` coordinate values
    x: Sequence[float]

    #: series of all `y` coordinate values
    y: Sequence[float]


class ECDF:
    """
    Three sets of coordinates for plotting an ECDF: inliers, outliers, and far
    outliers.
    """

    def __init__(self, inliers: XYSeries, outliers: XYSeries, far_outliers: XYSeries):
        """
        :param inliers: coordinates for inliers in the ECDF
        :param outliers: coordinates for outliers in the ECDF
        :param far_outliers: coordinates for far outliers in the ECDF
        """
        self._inliers = inliers
        self._outliers = outliers
        self._far_outliers = far_outliers

    @property
    def inliers(self) -> XYSeries:
        """
        Coordinates for inliers in the ECDF.
        """
        return self._inliers

    @property
    def outliers(self) -> XYSeries:
        """
        Coordinates for outliers in the ECDF.
        """
        return self._outliers

    @property
    def far_outliers(self) -> XYSeries:
        """
        Coordinates for far outliers in the ECDF.
        """
        return self._far_outliers


class ECDFStyle(DrawingStyle, metaclass=ABCMeta):
    """
    Base drawing style for ECDFs.
    """

    @abstractmethod
    def _draw_ecdf(
        self, ecdf: ECDF, x_label: str, iqr_multiple: float, iqr_multiple_far: float
    ) -> None:
        pass


__tracker.validate()



import logging
import re
from os import environ
from urllib import request
from xml.etree import ElementTree

from pytest import mark

import pytools

log = logging.getLogger(__name__)

ENV_RUN_PACKAGE_VERSION_TEST = "RUN_PACKAGE_VERSION_TEST"


@mark.skipif(
    condition=environ.get(ENV_RUN_PACKAGE_VERSION_TEST, "") != pytools.__name__,
    reason="Parent build is not primarily for pytools.",
)
def test_package_version() -> None:
    dev_version = pytools.__version__

    log.info(f"Test package version – version set to: {dev_version}")
    assert re.match(
        r"^(\d)+\.(\d)+\.(\d)+(rc\d+)?$", dev_version
    ), "pytools.__version__ is not in MAJOR.MINOR.PATCH[rcN] format."

    releases_uri = "https://pypi.org/rss/project/gamma-pytools/releases.xml"

    with request.urlopen(releases_uri) as response:
        assert response.getcode() == 200, "Error getting releases from PyPi"
        releases_xml = response.read()

    tree = ElementTree.fromstring(releases_xml)
    releases_nodes = tree.findall(path=".//channel//item//title")
    releases = [r.text for r in releases_nodes]

    log.info(f"Releases found on PyPi: {', '.join(releases)}")

    assert (
        dev_version not in releases
    ), f"Current package version {dev_version} already on PyPi"

    is_minor_or_major_release = dev_version.endswith(".0")

    if is_minor_or_major_release:
        pre_releases = [
            version
            for version in releases
            if re.match(f"{dev_version}rc\\d+$", version)
        ]

        assert pre_releases, (
            f"Release of major or minor version {dev_version} "
            f"requires at least one pre-release, e.g. {dev_version}rc0"
        )

        log.info(
            f"Pre-releases {pre_releases} exist(s) – "
            f"release of major/minor version {dev_version} allowed"
        )



"""
Basic test cases for the `pytools.api` module
"""
# noinspection PyPackageRequirements
import sys

import pytest

print(sys.path)

from pytools.api import deprecated


def test_deprecated() -> None:
    @deprecated
    def _f() -> None:
        pass

    with pytest.warns(
        expected_warning=FutureWarning,
        match="Call to deprecated function test_deprecated.<locals>._f",
    ):
        _f()

    @deprecated(message="test message")
    def _g() -> None:
        pass

    with pytest.warns(
        expected_warning=FutureWarning,
        match="Call to deprecated function test_deprecated.<locals>._g: test message",
    ):
        _g()



"""
Utilities related to matplotlib.
"""

import logging

from matplotlib.ticker import Formatter

from pytools.api import AllTracker
from pytools.meta import SingletonMeta

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = ["PercentageFormatter"]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


class PercentageFormatter(Formatter, metaclass=SingletonMeta):
    """
    Formats floats as a percentages with 3 digits precision, omitting trailing zeros.

    For percentages above 100%, formats percentages as the nearest whole number.

    Formatting examples:

    - ``0.00005`` is formatted as ``0.01%``
    - ``0.0005`` is formatted as ``0.05%``
    - ``0.0`` is formatted as ``0%``
    - ``0.1`` is formatted as ``10%``
    - ``1.0`` is formatted as ``100%``
    - ``0.01555`` is formatted as ``1.56%``
    - ``0.1555`` is formatted as ``15.6%``
    - ``1.555`` is formatted as ``156%``
    - ``15.55`` is formatted as ``1556%``
    - ``1555`` is formatted as ``1.6e+05%``
    """

    def __call__(self, x, pos=None) -> str:
        if x < 1.0:
            return f"{x * 100.0:.3g}%"
        else:
            return f"{round(x * 100.0):.5g}%"


# check consistency of __all__

__tracker.validate()



"""
Configuration file for the Sphinx documentation builder.

Receives majority of configuration from pytools conf_base.py
"""

import os
import sys

sys.path.insert(
    0,
    os.path.abspath(
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            os.pardir,
            os.pardir,
            os.pardir,
            "pytools",
            "sphinx",
            "base",
        )
    ),
)

from conf_base import set_config

# ----- custom configuration -----

set_config(
    globals(),
    project="pytools",
    modules=["pytools"],
    html_logo="_static/gamma_pytools_logo.png",
)



"""
Core implementation of :mod:`pytools.viz.matrix`.
"""

import logging
from typing import Any, Callable, Dict, Iterable, Optional, Tuple, Type, Union

import numpy as np
import pandas as pd
from matplotlib.axes import Axes, mticker
from matplotlib.axis import Axis
from matplotlib.colors import Normalize
from matplotlib.ticker import Formatter, FuncFormatter

from .. import ColorbarMatplotStyle, Drawer, TextStyle
from ..color import ColorScheme, text_contrast_color
from ..util import PercentageFormatter
from .base import MatrixStyle
from pytools.api import AllTracker, inheritdoc

log = logging.getLogger(__name__)


#
# exported names
#

__all__ = [
    "MatrixMatplotStyle",
    "PercentageMatrixMatplotStyle",
    "MatrixReportStyle",
    "MatrixDrawer",
]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


@inheritdoc(match="[see superclass]")
class MatrixMatplotStyle(MatrixStyle, ColorbarMatplotStyle):
    """
    Matplot style for matrices.

    Numerical values of matrix cells are rendered as colors from a color map,
    with a color bar attached as a legend.

    Matrix cells will be annotated with their values if a *cell format* is specified.
    Cell formats can be defined in different ways:

    - a new-style Python format string, e.g., ``{:.3g}``
    - a function accepting the value to be formatted as its only positional argument,
      e.g., ``lambda x: f"{x * 100:.3g}%"``
    - a :class:`~matplotlib.ticker.Formatter` instance
    """

    #: The maximum number of ticks to put on the x and y axis;
    #: ``None`` to determine the number of ticks automatically.
    max_ticks: Optional[Tuple[int, int]]

    #: Formatter for annotating each matrix cell with its value; if ``None``,
    #: no cells are annotated.
    cell_formatter: Optional[Formatter]

    def __init__(
        self,
        *,
        ax: Optional[Axes] = None,
        colors: Optional[ColorScheme] = None,
        colormap_normalize: Optional[Normalize] = None,
        colorbar_major_formatter: Optional[Formatter] = None,
        colorbar_minor_formatter: Optional[Formatter] = None,
        max_ticks: Optional[Tuple[int, int]] = None,
        cell_format: Union[str, Formatter, Callable[[Any], str], None] = None,
        **kwargs,
    ) -> None:
        """
        :param max_ticks: the maximum number of ticks to put on the x and y axis;
            ``None`` to determine the number of ticks automatically (default: ``None``)
        :param cell_format: format for annotating each matrix cell with
            its value if sufficient space is available (optional – do not annotate
            cells if omitted);
            if no colorbar major formatter is specified, use the cell format also
            as the colorbar major formatter
        """
        if cell_format is None:
            cell_formatter = None
        elif isinstance(cell_format, str):
            cell_formatter = FuncFormatter(func=lambda x, _: cell_format.format(x))
        elif isinstance(cell_format, Formatter):
            cell_formatter = cell_format
        elif callable(cell_format):
            cell_formatter = FuncFormatter(func=lambda x, _: cell_format(x))
        else:
            raise TypeError(
                "arg cell_format must be a format string, a Formatter, or a callable, "
                f"but a {type(cell_format).__name__} was passed"
            )

        if colorbar_major_formatter is None:
            colorbar_major_formatter = cell_formatter

        super().__init__(
            ax=ax,
            colors=colors,
            colormap_normalize=(
                colormap_normalize if colormap_normalize is not None else Normalize()
            ),
            colorbar_major_formatter=colorbar_major_formatter,
            colorbar_minor_formatter=colorbar_minor_formatter,
            **kwargs,
        )

        if max_ticks is not None and not (
            isinstance(max_ticks, Tuple)
            and len(max_ticks) == 2
            and all(isinstance(x, int) for x in max_ticks)
        ):
            raise ValueError(
                f"arg max_ticks={max_ticks} must be None or a tuple of 2 integers"
            )
        self.max_ticks = max_ticks
        self.cell_formatter = cell_formatter

    __init__.__doc__ = ColorbarMatplotStyle.__init__.__doc__ + __init__.__doc__

    def draw_matrix(self, matrix: pd.DataFrame) -> None:
        """[see superclass]"""
        ax: Axes = self.ax
        self.ax.margins(0, 0)

        # store values locally so we can label the matrix cells when finalizing the plot
        data = matrix.values
        ax.imshow(
            data,
            cmap=self.colors.colormap,
            norm=self.colormap_normalize,
            origin="upper",
            interpolation="nearest",
            aspect="equal",
        )

        # determine if a number of labels has been configured for this style
        max_ticks = self.max_ticks
        if max_ticks is None:
            max_x_ticks = max_y_ticks = None
        else:
            max_x_ticks, max_y_ticks = max_ticks

        # rotate x labels if they are categorical
        tick_params: Dict[bool, Dict[str, Any]] = {
            False: {},
            True: dict(rotation=45, ha="right"),
        }

        def _set_ticks(index: pd.Index, max_bins: int, axis: Axis, rotate: bool):
            # set the x and y ticks

            # determine number of bins
            if max_bins is not None:
                n_bins = max_bins
            elif index.is_numeric():
                n_bins = "auto"
            else:
                n_bins = len(index)

            locator = mticker.MaxNLocator(
                nbins=n_bins, steps=[1, 2, 5, 10], integer=True, prune="both"
            )
            axis.set_major_locator(locator)

            tick_locations: np.ndarray = axis.get_ticklocs()
            if len(index) > len(tick_locations):
                # we can plot only selected tick labels: look up labels for the
                # visible tick indices
                labels = index[tick_locations.astype(int)]
            else:
                # we can plot all tick labels
                labels = index.values

            # Replace the tick locator with a fixed locator, preserving the tick
            # locations determined by the MaxNLocator. This is needed for compatibility
            # with the FixedFormatter that will be created when setting the tick labels
            axis.set_ticks(axis.get_ticklocs())

            # Set the tick labels; behind the scenes this will create a FixedFormatter.
            axis.set_ticklabels(labels, **tick_params[rotate])

        _set_ticks(
            index=matrix.columns,
            max_bins=max_x_ticks,
            axis=ax.xaxis,
            rotate=not matrix.columns.is_numeric(),
        )
        _set_ticks(
            index=matrix.index, max_bins=max_y_ticks, axis=ax.yaxis, rotate=False
        )

        # get the matrix size
        n_rows = data.shape[0]
        n_columns = data.shape[1]

        # only draw labels if a cell formatter is defined, and minimal height/width
        # is available
        if self.cell_formatter is not None and all(
            size <= 1 for size in self.text_dimensions("0")
        ):
            # draw the axis to ensure we'll get correct coordinates
            ax.draw(self.renderer)

            # get the cell formatter as a local field
            cell_formatter = self.cell_formatter

            # render the text for every box where the text fits
            for y in range(n_rows):
                for x in range(n_columns):
                    x_text = x
                    y_text = y
                    cell_value = data[y, x]
                    label = cell_formatter(cell_value)
                    text_width, _ = self.text_dimensions(text=label, x=x_text, y=y_text)

                    if text_width > 1:
                        # show ellipsis in cells where the text does not fit
                        label = "…"

                    self.ax.text(
                        x=x_text,
                        y=y_text,
                        s=label,
                        ha="center",
                        va="center",
                        color=text_contrast_color(
                            bg_color=self.color_for_value(z=cell_value)
                        ),
                    )

        # hide spines
        for _, spine in ax.spines.items():
            spine.set_visible(False)

        # create a white grid using minor tick positions
        ax.set_xticks(np.arange(n_columns + 1) - 0.5, minor=True)
        ax.set_yticks(np.arange(n_rows + 1) - 0.5, minor=True)
        ax.grid(which="minor", color=self.colors.foreground, linestyle="-", linewidth=2)
        ax.tick_params(which="minor", bottom=False, left=False)


class PercentageMatrixMatplotStyle(MatrixMatplotStyle):
    """
    A matrix plot where all values are percentages.

    Annotates matrix cells with percentage values.
    """

    def __init__(
        self,
        *,
        ax: Optional[Axes] = None,
        colors: Optional[ColorScheme] = None,
        colormap_normalize: Optional[Normalize] = None,
        max_ticks: Optional[Tuple[int, int]] = None,
        **kwargs,
    ) -> None:
        """
        :param max_ticks: the maximum number of ticks to put on the x and y axis;
            ``None`` to determine the number of ticks automatically (default: ``None``)
        """
        _unsupported_fields = {
            "colorbar_major_formatter",
            "colorbar_minor_formatter",
            "cell_format",
        }.intersection(kwargs.keys())
        if _unsupported_fields:
            if len(_unsupported_fields) == 1:
                _args = f"arg {next(iter(_unsupported_fields))} is"
            else:
                _args = f'args {", ".join(_unsupported_fields)} are'
            raise ValueError(f"{_args} not supported by class {type(self).__name__}")

        super().__init__(
            ax=ax,
            colors=colors,
            colormap_normalize=(
                colormap_normalize
                if colormap_normalize
                else Normalize(vmin=0.0, vmax=1.0)
            ),
            max_ticks=max_ticks,
            colorbar_major_formatter=PercentageFormatter(),
            colorbar_minor_formatter=None,
            cell_format=lambda x: (
                f"{np.round(x * 100, 1):.2g}"
                if abs(x) < 0.1
                else f"{np.round(x * 100):.0f}"
            ),
            **kwargs,
        )

    __init__.__doc__ = ColorbarMatplotStyle.__init__.__doc__ + __init__.__doc__

    @classmethod
    def get_default_style_name(cls) -> str:
        """[see superclass]"""
        return f"{super().get_default_style_name()}%"


@inheritdoc(match="[see superclass]")
class MatrixReportStyle(MatrixStyle, TextStyle):
    """
    Text report style for matrices.
    """

    def draw_matrix(self, matrix: pd.DataFrame) -> None:
        """[see superclass]"""
        matrix.to_string(buf=self.out, line_width=self.width)


#
# Drawer classes
#


@inheritdoc(match="[see superclass]")
class MatrixDrawer(Drawer[pd.DataFrame, MatrixStyle]):
    """
    Drawer for matrices of numerical values.

    Supports the following pre-defined styles:

    - ``"matplot"``: `matplotlib` plot of the matrix using a
      :class:`.MatrixMatplotStyle`
    - ``"matplot%"``: `matplotlib` plot of matrix with annotations formatted as
      percentages, using a :class:`.PercentageMatrixMatplotStyle`
    - ``"text"``: text representation of the matrix, using a :class:`.MatrixReportStyle`
    """

    def __init__(self, style: Optional[Union[MatrixStyle, str]] = None) -> None:
        """
        :param style: the style to be used for drawing (default: ``"matplot"``)
        """
        super().__init__(style=style)

    @classmethod
    def get_style_classes(cls) -> Iterable[Type[MatrixStyle]]:
        """[see superclass]"""
        return [
            MatrixMatplotStyle,
            PercentageMatrixMatplotStyle,
            MatrixReportStyle,
        ]

    def _draw(self, data: pd.DataFrame) -> None:
        # draw the matrix
        self.style.draw_matrix(data)


__tracker.validate()



"""
Implementation of :mod:`pytools.expression` and subpackages.
"""

import logging
from typing import Any, Dict, Generic, TypeVar
from weakref import WeakValueDictionary

from ...api import AllTracker, inheritdoc
from ...meta import SingletonMeta, compose_meta
from ..base import AtomicExpression

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = [
    "Lit",
    "Id",
    "Epsilon",
]

#
# Type variables
#

T_Literal = TypeVar("T_Literal", bool, int, float, complex, str, bytes)


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Class definitions
#


@inheritdoc(match="[see superclass]")
class Lit(AtomicExpression[T_Literal], Generic[T_Literal]):
    """
    A literal value (usually a number or string).
    """

    def __init__(self, value: T_Literal):
        """
        :param value: the literal value represented by this expression
        """
        super().__init__()
        self._value = value

    @property
    def value_(self) -> T_Literal:
        """[see superclass]"""
        return self._value

    @property
    def text_(self) -> str:
        """[see superclass]"""
        return repr(self.value_)


class _IdentifierMeta(type):

    _identifiers: Dict[str, "Id"] = WeakValueDictionary()

    def __getattr__(self, name: str) -> "Id":
        if name.startswith("_") or name.endswith("_") or name == "Id":
            # we do not allow creating identifiers with leading or trailing underscores
            # we also disallow "Id" to avoid a compatibility issue with sphinx
            raise AttributeError(name)

        identifier = _IdentifierMeta._identifiers.get(name, None)
        if not identifier:
            _IdentifierMeta._identifiers[name] = identifier = Id(name)

        return identifier


@inheritdoc(match="[see superclass]")
class Id(
    AtomicExpression[str],
    metaclass=compose_meta(_IdentifierMeta, type(AtomicExpression)),
):
    """
    An identifier.

    Identifiers can be created either by

    - class instantiation: ``Id("x")``
    - attribute access: ``Id.x``

    The attribute access method will return the identical :class:`Id` instance
    for subsequent access to an attribute of the same name (stored internally using a
    weak reference dictionary).
    Attribute access requires attribute names that do not start or end with an
    underscore character (``_``), otherwise an :class:`AttributeError` is raised.
    For compatibility reasons, attribute access does not work for the ``Id`` name, i.e.,
    ``Id.Id`` will raise an :class:`AttributeError`.
    """

    def __init__(self, name: Any) -> None:
        """
        :param name: the name of the identifier
        """
        super().__init__()
        if not isinstance(name, str):
            name = getattr(name, "__name__", None)
            if not name:
                raise TypeError(
                    "arg name must be a string, or must have attribute __name__"
                )
        self._name = name

    @property
    def value_(self) -> str:
        """[see superclass]"""
        return self._name

    @property
    def text_(self) -> str:
        """[see superclass]"""
        return self._name


@inheritdoc(match="[see superclass]")
class Epsilon(
    AtomicExpression[None],
    metaclass=compose_meta(SingletonMeta, type(AtomicExpression)),
):
    """
    A singleton class representing the empty expression.
    """

    @property
    def value_(self) -> None:
        """[see superclass]"""
        return None

    @property
    def text_(self) -> str:
        """[see superclass]"""
        return ""


__tracker.validate()



"""
Drawing dendrograms.
"""

import logging
from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Type, Union

import numpy as np

from .. import Drawer
from ._linkage import LinkageTree
from ._style import DendrogramHeatmapStyle, DendrogramLineStyle, DendrogramReportStyle
from .base import DendrogramStyle, Node
from pytools.api import AllTracker, inheritdoc

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = ["DendrogramDrawer"]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


class _SubtreeInfo(NamedTuple):
    names: List[str]
    weight: float


@inheritdoc(match="[see superclass]")
class DendrogramDrawer(Drawer[LinkageTree, DendrogramStyle]):
    """
    Draws dendrogram representations of :class:`.LinkageTree` objects.
    """

    def __init__(self, style: Optional[Union[DendrogramStyle, str]] = None) -> None:
        """[see superclass]"""
        super().__init__(style=style)

    @classmethod
    def get_style_classes(cls) -> Iterable[Type[DendrogramStyle]]:
        """[see superclass]"""
        return [
            DendrogramHeatmapStyle,
            DendrogramLineStyle,
            DendrogramReportStyle,
        ]

    def _get_style_kwargs(self, data: LinkageTree) -> Dict[str, Any]:
        return dict(
            leaf_label=data.leaf_label,
            distance_label=data.distance_label,
            weight_label=data.weight_label,
            **super()._get_style_kwargs(data=data),
        )

    def _draw(self, data: LinkageTree) -> None:
        # draw the linkage tree

        node_weight = np.zeros(len(data), float)

        def _calculate_weights(n: Node) -> (float, int):
            # calculate the weight of a node and number of leaves below
            if n.is_leaf:
                weight = n.weight
                n_leaves = 1
            else:
                l, r = data.children(n)
                lw, ln = _calculate_weights(l)
                rw, rn = _calculate_weights(r)
                weight = lw + rw
                n_leaves = ln + rn
            node_weight[n.index] = weight / n_leaves
            return weight, n_leaves

        def _draw_node(node: Node, y: int, width: float) -> _SubtreeInfo:
            # Recursively draw the part of the dendrogram under a node.
            #
            # Arguments:
            # - node:  the node to be drawn
            # - y:     the value determining the position of the node with respect to
            #          the leaves of the tree
            # - width: width difference in the tree covered by the node
            #
            # Returns: _SubtreeInfo instance with labels and weights

            if node.is_leaf:
                self.style.draw_link_leg(
                    bottom=0.0,
                    top=width,
                    leaf=y,
                    weight=node.weight,
                    tree_height=data.max_distance,
                )

                return _SubtreeInfo(names=[node.name], weight=node.weight)

            else:
                child_left, child_right = data.children(node=node)
                if node_weight[child_left.index] > node_weight[child_right.index]:
                    child_left, child_right = child_right, child_left

                info_left = _draw_node(
                    node=child_left, y=y, width=node.children_distance
                )
                info_right = _draw_node(
                    node=child_right,
                    y=y + len(info_left.names),
                    width=node.children_distance,
                )

                info = _SubtreeInfo(
                    names=info_left.names + info_right.names,
                    weight=info_left.weight + info_right.weight,
                )

                self.style.draw_link_connector(
                    bottom=node.children_distance,
                    top=width,
                    first_leaf=y,
                    n_leaves_left=len(info_left.names),
                    n_leaves_right=len(info_right.names),
                    weight=info.weight,
                    tree_height=data.max_distance,
                )

                return info

        _calculate_weights(data.root)

        tree_info = _draw_node(node=data.root, y=0, width=data.max_distance)
        self.style.draw_leaf_names(names=tree_info.names)


__tracker.validate()



"""
Core implementation of :mod:`pytools.viz.colors`
"""

import logging
from typing import Set, Tuple, Union

from matplotlib import cm
from matplotlib.colors import Colormap, LinearSegmentedColormap, to_rgb

from pytools.api import AllTracker, inheritdoc, validate_element_types, validate_type
from pytools.expression import Expression, HasExpressionRepr
from pytools.expression.atomic import Id
from pytools.meta import SingletonMeta, compose_meta

log = logging.getLogger(__name__)

#
# Exported names
#

__all__ = [
    "RgbColor",
    "RgbaColor",
    "ColorScheme",
    "MatplotColorScheme",
    "FacetLightColorScheme",
    "FacetDarkColorScheme",
    "text_contrast_color",
]

#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())

#
# Type definitions
#

#: RGB color type for use in color schemas and colored drawing styles.
RgbColor = Tuple[float, float, float]

#: RGB + Alpha color type for use in color schemas and colored drawing styles.
RgbaColor = Tuple[float, float, float, float]

#
# Constants
#

#: Black.
_RGB_BLACK: RgbColor = to_rgb("black")

#: White.
_RGB_WHITE: RgbColor = to_rgb("white")

#: FACET light grey.
_RGB_LIGHT_GREY: RgbColor = to_rgb("#c8c8c8")

#: FACET grey.
_RGB_GREY: RgbColor = to_rgb("#9a9a9a")

#: FACET dark grey.
_RGB_DARK_GREY: RgbColor = to_rgb("#3d3a40")

#: FACET dark blue.
_RGB_DARK_BLUE: RgbColor = to_rgb("#295e7e")

#: FACET blue.
_RGB_LIGHT_BLUE: RgbColor = to_rgb("#30c1d7")

#: FACET green.
_RGB_LIGHT_GREEN: RgbColor = to_rgb("#43fda2")

#: FACET status green.
_RGB_GREEN: RgbColor = to_rgb("#3ead92")

#: FACET status amber.
_RGB_AMBER: RgbColor = to_rgb("#a8b21c")

#: FACET status red.
_RGB_RED: RgbColor = to_rgb("#e61c57")

#: FACET dark red.
_RGB_DARK_RED: RgbColor = to_rgb("#c41310")

#: Standard colormap for FACET.
_COLORMAP_FACET = LinearSegmentedColormap.from_list(
    name="facet",
    colors=[
        (0, _RGB_DARK_GREY),
        (0.25, _RGB_DARK_BLUE),
        (0.65, _RGB_LIGHT_BLUE),
        (1.0, _RGB_LIGHT_GREEN),
    ],
)


#
# Classes
#


@inheritdoc(match="[see superclass]")
class ColorScheme(HasExpressionRepr):
    """
    A color scheme mapping semantic color designations to RGB colors,
    allowing code to refer to colors by usage rather than specific RGB values.
    """

    def __init__(self, foreground: RgbColor, background: RgbColor, **colors: RgbColor):
        """
        :param foreground: the foreground color
        :param background: the background color
        :param colors: additional colors as keyword arguments; one or more of %COLORS%
        """
        validate_type(foreground, expected_type=tuple, name="arg foreground")
        validate_type(background, expected_type=tuple, name="arg background")
        unsupported_colors = colors.keys() - ColorScheme._SUPPORTED_COLORS
        if unsupported_colors:
            raise ValueError(f"unsupported color arguments: {unsupported_colors}")
        validate_element_types(colors.values(), expected_type=tuple, name="**colors")
        self._colors = {
            ColorScheme._COLOR_FOREGROUND: foreground,
            ColorScheme._COLOR_BACKGROUND: background,
            **colors,
        }

    @property
    def foreground(self) -> RgbColor:
        """
        The foreground color.
        """
        return self._colors[ColorScheme._COLOR_FOREGROUND]

    @property
    def background(self) -> RgbColor:
        """
        The background color.
        """
        return self._colors[ColorScheme._COLOR_BACKGROUND]

    @property
    def fill_1(self) -> RgbColor:
        """
        The primary fill color.

        Defaults to the halfway point between the foreground and background color
        if not defined explicitly.
        """
        return self._colors.get(ColorScheme._COLOR_FILL_1, None) or tuple(
            (f + b) / 2 for f, b in zip(self.foreground, self.background)
        )

    @property
    def fill_2(self) -> RgbColor:
        """
        The secondary fill color.

        Defaults to the primary fill color if not defined explicitly.
        """
        return self._colors.get(ColorScheme._COLOR_FILL_2, None) or self.fill_1

    @property
    def fill_3(self) -> RgbColor:
        """
        The tertiary fill color.

        Defaults to the primary fill color if not defined explicitly.
        """
        return self._colors.get(ColorScheme._COLOR_FILL_3, None) or self.fill_1

    @property
    def accent_1(self) -> RgbColor:
        """
        The primary accent color.

        Defaults to the foreground color if not defined explicitly.
        """
        return self._colors.get(ColorScheme._COLOR_ACCENT_1, None) or self.foreground

    @property
    def accent_2(self) -> RgbColor:
        """
        The secondary accent color.

        Defaults to the primary accent color if not defined explicitly.
        """
        return self._colors.get(ColorScheme._COLOR_ACCENT_2, None) or self.accent_1

    @property
    def accent_3(self) -> RgbColor:
        """
        The tertiary accent color.

        Defaults to the secondary accent color if not defined explicitly.
        """
        return self._colors.get(ColorScheme._COLOR_ACCENT_3, None) or self.accent_2

    @property
    def status_ok(self) -> RgbColor:
        """
        The traffic light color indicating "ok" status.

        Defaults to the primary accent color if not defined explicitly.
        """
        return self._colors.get(ColorScheme._COLOR_STATUS_OK, None) or self.accent_1

    @property
    def status_warning(self) -> RgbColor:
        """
        The traffic light color indicating "warning" status.

        Defaults to the secondary accent color if not defined explicitly.
        """
        return (
            self._colors.get(ColorScheme._COLOR_STATUS_WARNING, None) or self.accent_2
        )

    @property
    def status_critical(self) -> RgbColor:
        """
        The traffic light color indicating "critical" status.

        Defaults to the tertiary accent color if not defined explicitly.
        """
        return (
            self._colors.get(ColorScheme._COLOR_STATUS_CRITICAL, None) or self.accent_3
        )

    def to_expression(self) -> Expression:
        """[see superclass]"""
        return Id(type(self))(**self._colors)

    _COLOR_FOREGROUND = foreground.fget.__name__
    _COLOR_BACKGROUND = background.fget.__name__
    _COLOR_FILL_1 = fill_1.fget.__name__
    _COLOR_FILL_2 = fill_2.fget.__name__
    _COLOR_FILL_3 = fill_3.fget.__name__
    _COLOR_ACCENT_1 = accent_1.fget.__name__
    _COLOR_ACCENT_2 = accent_2.fget.__name__
    _COLOR_ACCENT_3 = accent_3.fget.__name__
    _COLOR_STATUS_OK = status_ok.fget.__name__
    _COLOR_STATUS_WARNING = status_warning.fget.__name__
    _COLOR_STATUS_CRITICAL = status_critical.fget.__name__

    _SUPPORTED_COLORS: Set[str]


ColorScheme._SUPPORTED_COLORS = {
    v
    for k, v in vars(ColorScheme).items()
    if k.startswith("_COLOR_")
    and (
        k
        not in {
            ColorScheme.foreground.fget.__name__,
            ColorScheme.background.fget.__name__,
        }
    )
}

# noinspection PyProtectedMember
ColorScheme.__init__.__doc__ = ColorScheme.__init__.__doc__.replace(
    "%COLORS%",
    ", ".join(sorted(map(lambda code: f"``{code}``", ColorScheme._SUPPORTED_COLORS))),
)


@inheritdoc(match="[see superclass]")
class MatplotColorScheme(ColorScheme):
    """
    A color scheme for use with `matplotlib`, based on *semantic* color designations
    and a default color bar, preventing code from referring to colors.

    Supports `matplot` standard colors codified as strings.

    For an overview of matplotlib's named color maps, see
    `here <https://matplotlib.org/tutorials/colors/colormaps.html>`_.
    """

    _colormap: Colormap

    def __init__(
        self,
        foreground: Union[RgbColor, str],
        background: Union[RgbColor, str],
        colormap: Union[Colormap, str],
        **colors,
    ):
        """
        :param colormap: the default colormap for this style
        """
        super().__init__(
            foreground=to_rgb(foreground),
            background=to_rgb(background),
            **{key: to_rgb(color) for key, color in colors.items()},
        )

        if isinstance(colormap, Colormap):
            self._colormap = colormap
        elif isinstance(colormap, str):
            self._colormap = cm.get_cmap(name=colormap)
        else:
            raise ValueError("arg colormap must be a Colormap or a string")

    __init__.__doc__ = ColorScheme.__init__.__doc__ + __init__.__doc__

    @property
    def colormap(self) -> Colormap:
        """
        The default colormap to use for color gradients.
        """
        return self._colormap

    def to_expression(self) -> Expression:
        """[see superclass]"""
        return Id(type(self))(**self._colors, colormap=self.colormap)


@inheritdoc(match="[see superclass]")
class _FacetColorScheme(MatplotColorScheme):
    def __init__(self, foreground: RgbColor, background: RgbColor) -> None:
        super().__init__(
            foreground=foreground,
            background=background,
            colormap=_COLORMAP_FACET,
            **{
                ColorScheme._COLOR_FILL_1: _RGB_LIGHT_GREY,
                ColorScheme._COLOR_FILL_2: _RGB_GREY,
                ColorScheme._COLOR_FILL_3: _RGB_DARK_GREY,
                ColorScheme._COLOR_ACCENT_1: _RGB_LIGHT_GREEN,
                ColorScheme._COLOR_ACCENT_2: _RGB_LIGHT_BLUE,
                ColorScheme._COLOR_ACCENT_3: _RGB_DARK_BLUE,
                ColorScheme._COLOR_STATUS_OK: _RGB_GREEN,
                ColorScheme._COLOR_STATUS_WARNING: _RGB_AMBER,
                ColorScheme._COLOR_STATUS_CRITICAL: _RGB_RED,
            },
        )

    def to_expression(self) -> Expression:
        """[see superclass]"""
        return Id(type(self))()


class FacetLightColorScheme(
    _FacetColorScheme, metaclass=compose_meta(SingletonMeta, type(_FacetColorScheme))
):
    """
    The default FACET color scheme with a light background.
    """

    def __init__(self) -> None:
        super().__init__(foreground=_RGB_BLACK, background=_RGB_WHITE)


class FacetDarkColorScheme(
    _FacetColorScheme, metaclass=compose_meta(SingletonMeta, type(_FacetColorScheme))
):
    """
    The default FACET color scheme with a dark background.
    """

    def __init__(self) -> None:
        super().__init__(foreground=_RGB_WHITE, background=_RGB_BLACK)


#
# Functions
#


def text_contrast_color(
    bg_color: Union[RgbColor, RgbaColor]
) -> Union[RgbColor, RgbaColor]:
    """
    Get a text color that maximises contrast with the given background color.

    Returns white for background luminance < 50%, and black otherwise.
    If an alpha channel is provided as the 4th element of the background color,
    it is included with the text color unchanged.

    :param bg_color: RGB encoded background color
    :return: the contrasting text color
    """
    fill_luminance = sum(bg_color[:3]) / 3
    text_color = _RGB_WHITE if fill_luminance < 0.5 else _RGB_BLACK
    if len(bg_color) > 3:
        text_color = (*text_color[:3], bg_color[3])
    return text_color


# check consistency of __all__

__tracker.validate()



"""
Linkage Tree.

:class:`LinkageTree` is the internal representation of dendrograms.

:class:`LinkageNode` and :class:`LeafNode` are the building blocks of
:class:`LinkageTree`. Both these classes inherit from :class:`BaseNode`.
"""

from typing import Any, Iterable, List, Optional, Sequence, Tuple

import numpy as np

from pytools.api import AllTracker, to_tuple
from pytools.viz.dendrogram.base import LeafNode, LinkageNode, Node

#
# Exported names
#

__all__ = ["LinkageTree"]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Class definitions
#


class LinkageTree:
    """
    A traversable tree derived from a SciPy linkage matrix.

    Supports :func:`len`, and numerical indexing and iteration of nodes.
    """

    __F_CHILD_LEFT = 0
    __F_CHILD_RIGHT = 1
    __F_CHILDREN_DISTANCE = 2
    __F_N_DESCENDANTS = 3

    #: The original linkage matrix created by :func:`scipy.cluster.hierarchy.linkage`.
    #:
    #: One row of the scipy linkage matrix is a quadruple:
    #: `(<index of left child>,
    #: <index of right child>,
    #: <distance between children>,
    #: <number of descendant nodes>)`,
    #: where the descendant nodes include the nodes from the entire sub-tree,
    #: from direct children down to leaf nodes.
    scipy_linkage_matrix: np.ndarray

    #: The maximum possible distance in the linkage tree; this determines the height of
    #: the tree to be drawn.
    max_distance: float

    #: A label describing the type/unit of distances
    #: passed in arg `scipy_linkage_matrix` (optional).
    distance_label: Optional[str]

    #: A label describing the type of names
    #: passed in arg `leaf_names` (optional).
    leaf_label: Optional[str]

    #: A label describing the type/unit of weights
    #: passed in arg `leaf_weights` (optional).
    weight_label: Optional[str]

    def __init__(
        self,
        *,
        scipy_linkage_matrix: np.ndarray,
        leaf_names: Iterable[str],
        leaf_weights: Iterable[float],
        max_distance: Optional[float] = None,
        distance_label: Optional[str] = None,
        leaf_label: Optional[str] = None,
        weight_label: Optional[str] = None,
    ) -> None:
        """
        :param scipy_linkage_matrix: linkage matrix calculated by function
            :func:`scipy.cluster.hierarchy.linkage`
        :param leaf_names: labels of the leaves
        :param leaf_weights: weight of the leaves; all values must range between `0.0`
            and `1.0`, and should add up to `1.0`
        :param max_distance: maximum theoretical distance value; this must be equal
            to, or greater than the maximum distance in arg `scipy_linkage_matrix`
            (optional)
        :param distance_label: a label describing the type and/or unit of distances
            passed in arg `scipy_linkage_matrix` (optional)
        :param leaf_label: a label describing the type of names
            passed in arg `leaf_names` (optional)
        :param weight_label: a label describing the type and/or unit of weights
            passed in arg `leaf_weights` (optional)
        """

        n_branches = len(scipy_linkage_matrix)
        n_leaves = n_branches + 1

        def _validate_leaves(var: Sequence[Any], var_name: str):
            if len(var) != n_leaves:
                raise ValueError(f"expected {n_leaves} values " f"for arg {var_name}")

        self.scipy_linkage_matrix = scipy_linkage_matrix

        leaf_names: Tuple[str, ...] = to_tuple(
            leaf_names, element_type=str, arg_name="leaf_names"
        )
        leaf_weights: Tuple[float, ...] = to_tuple(
            leaf_weights, element_type=float, arg_name="leaf_weights"
        )

        _validate_leaves(leaf_names, "leaf_labels")
        _validate_leaves(leaf_weights, "leaf_weights")

        if any(not (0.0 <= weight <= 1.0) for weight in leaf_weights):
            raise ValueError(
                "all values in arg leaf_weights are required to be in the range "
                "from 0.0 to 1.0"
            )

        self._nodes: List[Node] = [
            *[
                LeafNode(index=index, name=label, weight=weight)
                for index, (label, weight) in enumerate(zip(leaf_names, leaf_weights))
            ],
            *[
                LinkageNode(
                    index=index + n_leaves,
                    children_distance=scipy_linkage_matrix[index][
                        LinkageTree.__F_CHILDREN_DISTANCE
                    ],
                )
                for index in range(n_branches)
            ],
        ]

        root_children_distance = self._nodes[-1].children_distance
        if max_distance is None:
            max_distance = root_children_distance
        elif max_distance < root_children_distance:
            raise ValueError(
                f"arg max_distance={max_distance} must be equal to or greater than "
                f"the maximum distance (= {root_children_distance}) in the linkage tree"
            )

        self.max_distance = max_distance
        self.leaf_label = leaf_label
        self.weight_label = weight_label
        self.distance_label = distance_label

    @property
    def root(self) -> Node:
        """
        The root node of the linkage tree.
        """
        return self._nodes[-1]

    def children(self, node: Node) -> Optional[Tuple[Node, Node]]:
        """
        Get the children of the given node.

        :param node: the node for which to get the children
        :return: ``None`` if the node is a leaf, otherwise the pair of children
        """

        node_index = node.index
        nodes = self._nodes

        # check that the node is included in this tree
        if node_index >= len(nodes) or node is not nodes[node_index]:
            raise ValueError("arg node is not a node in this linkage tree")

        if node.is_leaf:
            return None
        else:
            # noinspection PyProtectedMember
            node_linkage = self.scipy_linkage_matrix[node_index - self.n_leaves]
            ix_c1, ix_c2 = node_linkage[
                [LinkageTree.__F_CHILD_LEFT, LinkageTree.__F_CHILD_RIGHT]
            ].astype(int)
            return nodes[ix_c1], nodes[ix_c2]

    @property
    def n_leaves(self) -> int:
        """
        The number of leave nodes in this linkage tree.
        """
        return len(self) - len(self.scipy_linkage_matrix)

    def __len__(self) -> int:
        return len(self._nodes)

    def __getitem__(self, item: int) -> Node:
        return self._nodes[item]


__tracker.validate()



"""
Implementation of :mod:`pytools.expression` and subpackages.
"""
import itertools
import logging
from abc import ABCMeta
from typing import Any, Tuple, TypeVar, Union, cast

from ...api import AllTracker, inheritdoc
from .. import Expression, make_expression
from ..atomic import Epsilon, Id
from ..base import (
    BracketPair,
    CollectionLiteral,
    InfixExpression,
    Invocation,
    Operation,
    SimplePrefixExpression,
    SingletonExpression,
)
from ..operator import BinaryOperator, Operator, UnaryOperator
from .base import DictEntry, KeywordArgument, LambdaDefinition

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = [
    "BracketedExpression",
    "ListLiteral",
    "TupleLiteral",
    "SetLiteral",
    "DictLiteral",
    "UnaryOperation",
    "BinaryOperation",
    "Call",
    "Index",
    "Lambda",
    "Attr",
]

#
# Type variables
#

T = TypeVar("T")
T_Literal = TypeVar("T_Literal", bool, int, float, complex, str, bytes)


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Class definitions
#


#
# Bracketed expressions
#


@inheritdoc(match="[see superclass]")
class BracketedExpression(SingletonExpression, metaclass=ABCMeta):
    """
    An expression surrounded by brackets.
    """

    def __init__(self, brackets: BracketPair, subexpression: Any) -> None:
        """
        :param brackets: the brackets enclosing this expression's subexpressions
        :param subexpression: the subexpression enclosed by brackets
        """
        super().__init__()
        self._brackets = brackets
        self._subexpression = make_expression(subexpression)

    @property
    def brackets_(self) -> BracketPair:
        """
        The brackets enclosing this expression's subexpressions.
        """
        return self._brackets

    @property
    def subexpression_(self) -> Expression:
        """
        The subexpression enclosed by the brackets.
        """
        return self._subexpression

    @property
    def precedence_(self) -> int:
        """[see superclass]"""
        return Operator.MAX_PRECEDENCE

    def hash_(self) -> int:
        """[see superclass]"""
        return hash((type(self), self.brackets_, self.subexpression_.hash_()))

    def _eq_same_type(self, other: "BracketedExpression") -> bool:
        return self.brackets_ == other.brackets_ and self.subexpression_.eq_(
            other.subexpression_
        )


class ListLiteral(CollectionLiteral):
    """
    A list expression.
    """

    def __init__(self, *elements: Any):
        """
        :param elements: the list elements
        """
        super().__init__(brackets=BracketPair.SQUARE, elements=elements)


class TupleLiteral(CollectionLiteral):
    """
    A tuple expression.
    """

    def __init__(self, *elements: Any):
        """
        :param elements: the tuple elements
        """
        super().__init__(brackets=BracketPair.ROUND, elements=elements)


class SetLiteral(CollectionLiteral):
    """
    A set expression.
    """

    def __init__(self, *elements: Any):
        """
        :param elements: the set elements
        """
        super().__init__(brackets=BracketPair.CURLY, elements=elements)


class DictLiteral(CollectionLiteral):
    """
    A dictionary expression.
    """

    def __init__(self, *args: Tuple[Any, Any], **kwargs: Tuple[str, Any]):
        """
        :param args: dictionary entries as tuples ``(key, value)``
        :param kwargs: dictionary entries as keyword arguments
        """
        super().__init__(
            brackets=BracketPair.CURLY,
            elements=(
                DictEntry(key, value)
                for key, value in itertools.chain(args, kwargs.items())
            ),
        )


#
# Operations
#


@inheritdoc(match="[see superclass]")
class UnaryOperation(SimplePrefixExpression, Operation, metaclass=ABCMeta):
    """
    A unary operation.
    """

    def __init__(self, operator: UnaryOperator, operand: Any):
        super().__init__(prefix=Epsilon(), body=operand)
        self._operator = operator

    @property
    def separator_(self) -> str:
        """[see superclass]"""
        return self._operator.symbol

    @property
    def operator_(self) -> UnaryOperator:
        """[see superclass]"""
        return self._operator

    @property
    def operands_(self) -> Tuple[Expression, ...]:
        """[see superclass]"""
        return self.subexpressions_

    @property
    def precedence_(self) -> int:
        """[see superclass]"""
        return self._operator.precedence


@inheritdoc(match="[see superclass]")
class BinaryOperation(InfixExpression, Operation):
    """
    A operation with two or more operands.
    """

    def __init__(self, operator: BinaryOperator, *operands: Any):
        """
        :param operator: the binary operator applied by this operation
        :param operands: the operands applied to the binary operator
        """
        super().__init__()
        if not isinstance(operator, BinaryOperator):
            raise TypeError(f"arg operator={operator} must be an BinaryOperator")

        if len(operands) < 2:
            raise ValueError("operation requires at least two operands")

        operands: Tuple[Expression, ...] = tuple(
            make_expression(operand) for operand in operands
        )

        first_operand = operands[0]

        if isinstance(first_operand, BinaryOperation):
            if first_operand.operator_ == operator:
                # if first operand has the same operator, we flatten the operand
                # noinspection PyUnresolvedReferences
                operands = (*first_operand.operands_, *operands[1:])

        self._operator = operator
        self._operands = operands

    @property
    def operator_(self) -> BinaryOperator:
        """[see superclass]"""
        return self._operator

    @property
    def operands_(self) -> Tuple[Expression, ...]:
        """[see superclass]"""
        return self._operands

    @property
    def infix_(self) -> BinaryOperator:
        """[see superclass]"""
        return self._operator

    @property
    def precedence_(self) -> int:
        """[see superclass]"""
        return self.infix_.precedence


#
# Invocations
#


class Call(Invocation):
    """
    A call expression.
    """

    def __init__(self, callee: Any, *args: Any, **kwargs: Any):
        """
        :param callee: the expression representing the object being called
        :param args: the positional argument(s) of the call
        :param kwargs: the keyword arguments of the call
        """
        super().__init__(
            prefix=callee,
            brackets=BracketPair.ROUND,
            args=(
                *args,
                *(KeywordArgument(name, item) for name, item in kwargs.items()),
            ),
        )

    @property
    def callee_(self) -> Expression:
        """
        The expression being called; identical with the prefix of this expression.
        """
        return self.prefix_


class Index(Invocation):
    """
    An indexing operation.
    """

    def __init__(self, collection: Any, key: Any):
        """
        :param collection: the collection to be indexed
        :param key: the index key
        """
        keys = key if isinstance(key, tuple) else (key,)
        super().__init__(prefix=collection, brackets=BracketPair.SQUARE, args=keys)

    @property
    def collection_(self) -> Expression:
        """
        The collection to be indexed; same as :attr:`prefix_`.
        """
        return self.prefix_


#
# Lambda
#


@inheritdoc(match="[see superclass]")
class Lambda(SimplePrefixExpression):
    """
    A lambda expression.
    """

    _PRECEDENCE = UnaryOperator.LAMBDA.precedence

    def __init__(self, *params: Id, body: Any):
        """
        :param params: the parameters of the lambda expression
        :param body: the body of the lambda expression
        """
        super().__init__(prefix=Epsilon(), body=LambdaDefinition(*params, body=body))
        self._has_params = len(params) > 0

    @property
    def precedence_(self) -> int:
        """[see superclass]"""
        return Lambda._PRECEDENCE

    @property
    def separator_(self) -> str:
        """[see superclass]"""
        if self._has_params:
            return f"{UnaryOperator.LAMBDA.symbol} "
        else:
            return UnaryOperator.LAMBDA.symbol


#
# Attr
#


class Attr(BinaryOperation):
    """
    The ``….…`` ("dot") operation to reference an attribute of an object.
    """

    def __init__(self, obj: Any, attribute: Union[Id, str]) -> None:
        """
        :param obj: the object whose attribute is referenced
        :param attribute: the name of the attribute being referenced
        """
        if isinstance(attribute, str):
            attribute = Id(attribute)
        elif not isinstance(attribute, cast(type, Id)):
            raise TypeError("arg attribute must be a string or an Identifier")

        if isinstance(obj, Attr):
            sub = obj.subexpressions_
            super().__init__(BinaryOperator.DOT, sub[0], *(sub[1:]), attribute)
        else:
            super().__init__(BinaryOperator.DOT, obj, attribute)


__tracker.validate()



"""
String representations of expressions.
"""

import logging
from abc import ABCMeta, abstractmethod
from typing import List, NamedTuple, Tuple

from .. import Expression, ExpressionAlias, ExpressionFormatter
from ..base import (
    AtomicExpression,
    BracketedExpression,
    BracketPair,
    InfixExpression,
    PrefixExpression,
)
from ..operator import BinaryOperator
from pytools.api import AllTracker

log = logging.getLogger(__name__)

__all__ = ["PythonExpressionFormatter"]


#
# Private helper classes
#


class FormattingConfig(NamedTuple):
    """
    The parameters to use for formatting an expression.
    """

    max_width: int = 80
    """maximum line width"""
    indent_width: int = 4
    """number of spaces in one indentation"""
    single_line: bool = False
    """if ``True``, always produce a single line regardless of width"""


class IndentedLine(NamedTuple):
    """
    An indented line of text.
    """

    #: the number of indentation levels
    indent: int
    #: the text in this line
    text: str

    def prepend(self, text: str) -> "IndentedLine":
        """
        Add the given text to the start of this indented line.

        :param text: the text to add
        :return: a copy of this indented line, with the text added
        """
        return IndentedLine(indent=self.indent, text=text + self.text)

    def append(self, text: str) -> "IndentedLine":
        """
        Add the given text to the end of this indented line.

        :param text: the text to add
        :return: a copy of this indented line, with the text added
        """
        return IndentedLine(indent=self.indent, text=self.text + text)

    def __add__(self, other: str) -> "IndentedLine":
        return self.append(text=other)

    def __radd__(self, other: str) -> "IndentedLine":
        return self.prepend(text=other)

    def __len__(self) -> int:
        return len(self.text)


class TextualForm:
    """
    A hierarchical textual representation of an expression.
    """

    #: the default formatting configuration for rendering textual forms
    DEFAULT_FORMAT = FormattingConfig()

    @staticmethod
    def from_expression(expression: Expression) -> "TextualForm":
        """
        Generate a textual form for the given expression.

        :param expression: the expression to be transformed
        :return: the resulting textual form
        """

        from ..atomic import Epsilon

        if expression is Epsilon():
            return EMPTY_FORM
        if isinstance(expression, AtomicExpression):
            return AtomicForm(expression)
        elif isinstance(expression, BracketedExpression):
            return BracketedForm.from_bracketed_expression(expression)
        elif isinstance(expression, InfixExpression):
            return InfixForm.from_infix_expression(expression)
        elif isinstance(expression, PrefixExpression):
            return PrefixForm.from_prefix_expression(expression)
        elif isinstance(expression, ExpressionAlias):
            return TextualForm.from_expression(expression.expression_)
        else:
            raise TypeError(f"unknown expression type: {type(expression)}")

    @property
    def needs_multi_line_encapsulation(self) -> bool:
        """
        If ``True``, this form should be encapsulated in brackets when rendered
        across multiple lines.
        """
        return False

    def to_text(self, config: FormattingConfig = DEFAULT_FORMAT) -> str:
        """
        Render this textual form as a string.

        :param config: the formatting configuration to use
        :return: the resulting string
        """

        if config.single_line:

            return self.to_single_line()

        else:

            def _spacing(indent: int) -> str:
                return " " * (config.indent_width * indent)

            return "\n".join(
                f"{_spacing(indent)}{text}"
                for indent, text in self.to_lines(config=config)
            )

    @abstractmethod
    def to_lines(
        self,
        config: FormattingConfig,
        indent: int = 0,
        leading_characters: int = 0,
        trailing_characters: int = 0,
    ) -> List[IndentedLine]:
        """
        Generate a list of indented lines from this textual form.

        :param config: the rendering configuration
        :param indent: the indentation level to use as a starting point
        :param leading_characters: space to reserve in the first line for leading
            characters (needed to determine whether maximum width has been exceeded)
        :param trailing_characters: space to reserve in the last line for trailing
            characters (needed to determine whether maximum width has been exceeded)
        :return: a list of indented lines generated from this textual form
        """
        pass

    @abstractmethod
    def to_single_line(self) -> str:
        """
        Convert this representation to a single-line string.

        :return: the resulting string
        """
        pass

    def encapsulate(
        self, *, condition: bool = True, single_line: bool = True
    ) -> "TextualForm":
        """
        Return this form encapsulated in round parentheses.

        :param condition: if ``False``, do not encapsulate this form
        :param single_line: if ``False``, render the encapsulation only when the form
            is rendered across multiple lines
        :return: the resulting form depending on the condition
        """
        return (
            BracketedForm(
                brackets=BracketPair.ROUND, subform=self, single_line=single_line
            )
            if condition
            else self
        )

    @abstractmethod
    def __len__(self) -> int:
        pass

    def __repr__(self) -> str:
        return self.to_text()


class EmptyForm(TextualForm):
    """
    An empty form representing the _epsilon_ expression.
    """

    def to_lines(
        self,
        config: FormattingConfig,
        indent: int = 0,
        leading_characters: int = 0,
        trailing_characters: int = 0,
    ) -> List[IndentedLine]:
        """[see superclass]"""
        return []

    to_lines.__doc__ = TextualForm.to_lines.__doc__

    def to_single_line(self) -> str:
        """[see superclass]"""
        return ""

    to_single_line.__doc__ = TextualForm.to_single_line.__doc__

    def __len__(self) -> int:
        """[see superclass]"""
        return 0


EMPTY_FORM = EmptyForm()


class AtomicForm(TextualForm):
    """
    A textual representation of an atomic expression.
    """

    def __init__(self, expression: AtomicExpression) -> None:
        self.text = expression.text_

    def to_lines(
        self,
        config: FormattingConfig,
        indent: int = 0,
        leading_characters: int = 0,
        trailing_characters: int = 0,
    ) -> List[IndentedLine]:
        """[see superclass]"""

        return [IndentedLine(indent=indent, text=self.to_single_line())]

    to_lines.__doc__ = TextualForm.to_lines.__doc__

    def to_single_line(self) -> str:
        """[see superclass]"""

        return self.text

    to_single_line.__doc__ = TextualForm.to_single_line.__doc__

    def __len__(self) -> int:
        return len(self.text)


class ComplexForm(TextualForm, metaclass=ABCMeta):
    """
    Base class of non-atomic forms.
    """

    def __init__(self, length: int) -> None:
        """
        :param length: the length of the single-line representation of this form
        """

        self._len = length

    def to_lines(
        self,
        config: FormattingConfig,
        indent: int = 0,
        leading_characters: int = 0,
        trailing_characters: int = 0,
    ) -> List[IndentedLine]:
        """[see superclass]"""

        if (
            leading_characters
            + len(self)
            + indent * config.indent_width
            + trailing_characters
            > config.max_width
        ):
            return self.to_multiple_lines(
                config=config,
                indent=indent,
                leading_characters=leading_characters,
                trailing_characters=trailing_characters,
            )
        else:
            return [IndentedLine(indent=indent, text=self.to_single_line())]

    @abstractmethod
    def to_multiple_lines(
        self,
        config: FormattingConfig,
        indent: int,
        leading_characters: int,
        trailing_characters: int,
    ) -> List[IndentedLine]:
        """
        Convert this representation to multiple lines.

        :param config: the formatting configuration to use
        :param indent: global indent of this expression
        :param leading_characters: leading space to reserve in first line
        :param trailing_characters: trailing space to reserve in last line
        :return: resulting lines
        """
        pass

    def __len__(self) -> int:
        return self._len


class BracketedForm(ComplexForm):
    """
    A hierarchical textual representation of a complex expression.
    """

    def __init__(
        self, brackets: BracketPair, subform: TextualForm, single_line: bool = True
    ) -> None:
        """
        :param brackets: the brackets surrounding the subform(s)
        :param subform: the subform to be bracketed
        :param single_line: if ``False``, do not render the brackets in single-line
            output
        """

        super().__init__(
            length=(
                (len(brackets.opening) + len(brackets.closing)) if single_line else 0
            )
            + len(subform)
        )

        self.brackets = brackets
        self.subform = subform
        self.single_line = single_line

    @staticmethod
    def from_bracketed_expression(expression: BracketedExpression) -> "BracketedForm":
        """
        Make a bracketed from for the given bracketed expression.

        :param expression: the bracketed expression to convert
        :return: the resulting bracketed form
        """
        return BracketedForm(
            brackets=expression.brackets_,
            subform=TextualForm.from_expression(expression.subexpression_),
        )

    def to_single_line(self) -> str:
        """[see superclass]"""
        subform_text = self.subform.to_single_line()
        if self.single_line:
            # render the brackets only when they are visible in single-line forms
            return f"{self.brackets.opening}{subform_text}{self.brackets.closing}"
        else:
            return subform_text

    to_single_line.__doc__ = TextualForm.to_single_line.__doc__

    def to_multiple_lines(
        self,
        config: FormattingConfig,
        indent: int,
        leading_characters: int,
        trailing_characters: int,
    ) -> List[IndentedLine]:
        """[see superclass]"""

        return [
            IndentedLine(indent=indent, text=self.brackets.opening),
            *self.subform.to_lines(
                config=config,
                indent=indent + 1,
                leading_characters=0,
                trailing_characters=0,
            ),
            IndentedLine(indent=indent, text=self.brackets.closing),
        ]

    to_multiple_lines.__doc__ = ComplexForm.to_multiple_lines.__doc__


class PrefixForm(ComplexForm):
    """
    A hierarchical textual representation of a complex expression.
    """

    def __init__(self, prefix: TextualForm, separator: str, body: TextualForm) -> None:
        """
        :param prefix: the prefix form
        :param separator: characters separating the prefix from the subform
        :param body: the body form
        """
        super().__init__(length=len(prefix) + len(separator) + len(body))

        self.prefix = prefix
        self.separator = separator
        self.body = body

    @staticmethod
    def from_prefix_expression(expression: PrefixExpression) -> TextualForm:
        """
        Create a prefixed form from the given prefix expression.
        """

        prefix = expression.prefix_
        prefix_form = TextualForm.from_expression(prefix).encapsulate(
            condition=prefix.precedence_ < expression.precedence_
        )

        body = expression.body_
        encapsulate_body = body.precedence_ < expression.precedence_
        body_form = TextualForm.from_expression(body)

        body_form = body_form.encapsulate(
            condition=encapsulate_body or body_form.needs_multi_line_encapsulation,
            single_line=encapsulate_body,
        )

        separator = expression.separator_
        if len(prefix_form) and separator[:1].isalpha():
            separator = " " + separator
        if len(body_form) and separator[-1:].isalpha():
            separator += " "

        return PrefixForm(prefix=prefix_form, separator=separator, body=body_form)

    @property
    def needs_multi_line_encapsulation(self) -> bool:
        """[see superclass]"""
        return self.prefix.needs_multi_line_encapsulation

    needs_multi_line_encapsulation.__doc__ = (
        TextualForm.needs_multi_line_encapsulation.__doc__
    )

    def to_single_line(self) -> str:
        """[see superclass]"""

        return (
            self.prefix.to_single_line() + self.separator + self.body.to_single_line()
        )

    def to_multiple_lines(
        self,
        config: FormattingConfig,
        indent: int,
        leading_characters: int,
        trailing_characters: int,
    ) -> List[IndentedLine]:
        """[see superclass]"""

        prefix_lines: List[IndentedLine] = self.prefix.to_lines(
            config=config,
            indent=indent,
            leading_characters=leading_characters,
            trailing_characters=0,
        )

        separator = self.separator

        subform_lines = self.body.to_lines(
            config=config,
            indent=indent,
            leading_characters=(len(prefix_lines[-1]) if prefix_lines else 0)
            + len(separator),
            trailing_characters=trailing_characters,
        )

        if prefix_lines:
            subform_lines[0] = prefix_lines[-1] + separator + subform_lines[0].text
        else:
            subform_lines[0] = separator + subform_lines[0]

        return prefix_lines[:-1] + subform_lines

    to_multiple_lines.__doc__ = ComplexForm.to_multiple_lines.__doc__


class InfixForm(ComplexForm):
    """
    A hierarchical textual representation of a complex expression.
    """

    PADDING_NONE = "none"
    PADDING_RIGHT = "right"
    PADDING_BOTH = "both"

    __PADDING_SPACES = {PADDING_NONE: 0, PADDING_RIGHT: 1, PADDING_BOTH: 2}

    def __init__(
        self,
        infix: str = "",
        infix_padding: str = PADDING_BOTH,
        subforms: Tuple[TextualForm, ...] = (),
    ) -> None:
        """
        :param infix: the infix operator separating the subforms
        :param infix_padding: the padding mode for the infix operator
        :param subforms: the subforms
        """

        if infix_padding is InfixForm.PADDING_RIGHT:
            subforms = tuple(
                subform.encapsulate(
                    condition=isinstance(subform, InfixForm), single_line=False
                )
                for subform in subforms
            )

        super().__init__(
            length=(
                sum(len(inner_representation) for inner_representation in subforms)
                + max(len(subforms) - 1, 0)
                * (len(infix) + (InfixForm.__PADDING_SPACES[infix_padding]))
            )
        )

        self.infix = infix
        self.infix_padding = infix_padding
        self.subforms = subforms

    @staticmethod
    def from_infix_expression(expression: InfixExpression) -> TextualForm:
        """
        Create a infix form from the given infix expression.

        :param expression: the infix expression to render the text from
        :return: the resulting textual form
        """

        from ..atomic import Epsilon

        subexpressions = expression.subexpressions_
        n_subexpressions = len(subexpressions)
        if n_subexpressions == 1:
            return TextualForm.from_expression(subexpressions[0])

        last_subexpression = n_subexpressions - 1
        subforms = tuple(
            TextualForm.from_expression(subexpression).encapsulate(
                condition=(
                    subexpression.precedence_ < expression.precedence_
                    if pos == 0
                    else (
                        subexpression.precedence_ <= expression.precedence_
                        and (
                            # a trailing prefix expression with an empty prefix does
                            # not need to be encapsulated (e.g., unary operators)
                            pos < last_subexpression
                            or not (
                                isinstance(subexpression, PrefixExpression)
                                and subexpression.prefix_ is Epsilon()
                            )
                        )
                    )
                )
            )
            for pos, subexpression in enumerate(subexpressions)
        )

        infix = expression.infix_

        infix_padding = (
            InfixForm.PADDING_RIGHT
            if infix in [BinaryOperator.COMMA, BinaryOperator.COLON]
            else InfixForm.PADDING_NONE
            if infix in [BinaryOperator.DOT, BinaryOperator.SLICE, BinaryOperator.NONE]
            else InfixForm.PADDING_BOTH
        )

        return InfixForm(
            infix=infix.symbol, infix_padding=infix_padding, subforms=subforms
        )

    @property
    def needs_multi_line_encapsulation(self) -> bool:
        """[see superclass]"""
        return True

    needs_multi_line_encapsulation.__doc__ = (
        TextualForm.needs_multi_line_encapsulation.__doc__
    )

    def to_single_line(self) -> str:
        """[see superclass]"""

        if self.infix:
            infix_padding = self.infix_padding
            if infix_padding is InfixForm.PADDING_NONE:
                infix = self.infix
            elif infix_padding is InfixForm.PADDING_RIGHT:
                infix = f"{self.infix} "
            else:
                assert (
                    infix_padding is InfixForm.PADDING_BOTH
                ), "known infix padding mode"
                infix = f" {self.infix} "
        else:
            infix = ""

        return infix.join(subform.to_single_line() for subform in self.subforms)

    to_single_line.__doc__ = TextualForm.to_single_line.__doc__

    def to_multiple_lines(
        self,
        config: FormattingConfig,
        indent: int,
        leading_characters: int,
        trailing_characters: int,
    ) -> List[IndentedLine]:
        """[see superclass]"""

        subforms: Tuple[TextualForm, ...] = self.subforms

        result: List[IndentedLine] = []

        if len(subforms) == 1:

            result.extend(
                subforms[0].to_lines(
                    config=config,
                    indent=indent,
                    leading_characters=leading_characters,
                    trailing_characters=trailing_characters,
                )
            )

        else:

            last_idx = len(subforms) - 1
            infix = self.infix

            if self.infix_padding is InfixForm.PADDING_RIGHT:
                len_infix = len(infix)
                for idx, inner_representation in enumerate(subforms):
                    lines = inner_representation.to_lines(
                        config=config,
                        indent=indent,
                        leading_characters=(leading_characters if idx == 0 else 0),
                        trailing_characters=(
                            len_infix if idx < last_idx else trailing_characters
                        ),
                    )

                    if idx != last_idx:
                        # append infix to last line,
                        # except we're in the last representation
                        lines[-1] += infix

                    result.extend(lines)
            else:
                if self.infix_padding is InfixForm.PADDING_BOTH:
                    infix += " "

                len_infix = len(infix)
                for idx, inner_representation in enumerate(subforms):
                    lines: List[IndentedLine] = inner_representation.to_lines(
                        config=config,
                        indent=indent,
                        leading_characters=leading_characters
                        if idx == 0
                        else len_infix,
                        trailing_characters=(
                            trailing_characters if idx == last_idx else 0
                        ),
                    )
                    if idx != 0:
                        # prepend infix to first line,
                        # except we're in the first representation
                        if lines:
                            lines[0] = infix + lines[0]
                        else:
                            lines = [IndentedLine(indent=indent, text=infix)]

                    result.extend(lines)

        return result

    to_multiple_lines.__doc__ = ComplexForm.to_multiple_lines.__doc__


#
# Public classes
#

__tracker = AllTracker(globals())


class PythonExpressionFormatter(ExpressionFormatter):
    """
    Formats expression objects as Python expressions, in line with the `black` style.
    """

    def __init__(
        self, max_width: int = 80, indent_width: int = 4, single_line: bool = False
    ):
        """
        :param max_width: the maximum line width (ignored when enforcing single-line
            text (default: 80)
        :param indent_width: the width of one indentation in spaces (default: 4)
        :param single_line: if ``False``, include line breaks to keep the width within
            maximum bounds (default: ``False``)
        """
        self._config = FormattingConfig(
            max_width=max_width, indent_width=indent_width, single_line=single_line
        )

    def to_text(self, expression: Expression) -> str:
        """[see superclass]"""

        form = TextualForm.from_expression(expression)

        # if the outermost form needs multiline encapsulation, do it
        form = form.encapsulate(
            condition=form.needs_multi_line_encapsulation, single_line=False
        )

        return form.to_text(self._config)

    to_text.__doc__ = ExpressionFormatter.to_text.__doc__


__tracker.validate()



#!/usr/bin/env python3
"""
Facet build script wrapping conda-build, and exposing matrix
dependency definition of pyproject.toml as environment variables
"""
import importlib
import importlib.util
import os
import pathlib
import re
import shutil
import subprocess
import sys
import warnings
from abc import ABCMeta, abstractmethod
from glob import glob
from typing import Any, Dict, List
from urllib.request import pathname2url

import toml

ALL_PROJECTS_QUALIFIER = "all"

CWD = os.getcwd()
SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))

FACET_PATH_ENV = "FACET_PATH"
FACET_PATH_URI_ENV = "FACET_PATH_URI"
FACET_BUILD_PKG_VERSION_ENV = "FACET_BUILD_{project}_VERSION"
CONDA_BUILD_PATH_ENV = "CONDA_BLD_PATH"

# pyproject.toml: elements of the hierarchy
TOML_BUILD = "build"
TOML_DIST_NAME = "dist-name"
TOML_FLIT = "flit"
TOML_MATRIX = "matrix"
TOML_METADATA = "metadata"
TOML_REQUIRES = "requires"
TOML_REQUIRES_PYTHON = "requires-python"
TOML_TOOL = "tool"


PROJ_PYTOOLS = "pytools"
PROJ_FACET = "facet"
PROJ_FLOW = "flow"
PROJ_SKLEARNDF = "sklearndf"
KNOWN_PROJECTS = {PROJ_PYTOOLS, PROJ_SKLEARNDF, PROJ_FACET, PROJ_FLOW}

B_CONDA = "conda"
B_TOX = "tox"
KNOWN_BUILD_SYSTEMS = {B_CONDA, B_TOX}

DEFAULT_DEPS = "default"
MIN_DEPS = "min"
MAX_DEPS = "max"
KNOWN_DEPENDENCY_TYPES = {DEFAULT_DEPS, MIN_DEPS, MAX_DEPS}

CONDA_BUILD_PATH_SUFFIX = os.path.join("dist", "conda")
TOX_BUILD_PATH_SUFFIX = os.path.join("dist", "tox")


class Builder(metaclass=ABCMeta):
    def __init__(self, project: str, dependency_type: str):
        self.project = project
        self.dependency_type = dependency_type

        # determine the FACET root path containing the project working directories

        if (FACET_PATH_ENV in os.environ) and os.environ[FACET_PATH_ENV]:
            facet_path = os.environ[FACET_PATH_ENV]
        else:
            facet_path = os.path.abspath(os.path.join(SCRIPT_DIR, os.pardir))
            os.environ[FACET_PATH_ENV] = facet_path

        if " " in facet_path:
            warnings.warn(
                f"The build base path '{facet_path}' contains spaces – "
                f"this causes issues with conda-build. "
                f"Consider to set a different path using the "
                f"environment variable {FACET_PATH_ENV} ahead of running make.py."
            )
        os.environ[FACET_PATH_URI_ENV] = f"file://{pathname2url(facet_path)}"

        # determine the package version of the project

        project_src = os.path.abspath(os.path.join(facet_path, project, "src"))
        if project in (PROJ_SKLEARNDF, PROJ_FLOW):
            # For sklearndf and flow,
            # __init__ can't be trivially imported due to import dependencies.
            # Load the version as defined in project._version module.
            spec = importlib.util.spec_from_file_location(
                "_version", os.path.join(project_src, project, "_version.py")
            )
        else:
            # pytools/facet: retrieve version from __init__.py
            spec = importlib.util.spec_from_file_location(
                "_version", os.path.join(project_src, project, "__init__.py")
            )
        version_module = importlib.util.module_from_spec(spec)
        # noinspection PyUnresolvedReferences
        spec.loader.exec_module(version_module)
        # noinspection PyUnresolvedReferences
        package_version = version_module.__version__

        os.environ[
            FACET_BUILD_PKG_VERSION_ENV.format(project=project.upper())
        ] = package_version

        self.package_version = package_version

    @staticmethod
    def for_build_system(
        build_system: str, project: str, dependency_type: str
    ) -> "Builder":
        if build_system == B_CONDA:
            return CondaBuilder(project=project, dependency_type=dependency_type)
        elif build_system == B_TOX:
            return ToxBuilder(project=project, dependency_type=dependency_type)
        else:
            raise KeyError(f"Unknown build system: {build_system}")

    @property
    @abstractmethod
    def build_system(self) -> str:
        pass

    @property
    @abstractmethod
    def build_path_suffix(self) -> str:
        pass

    def make_build_path(self, project: str = None) -> str:
        """
        Return the target build path for Conda or Tox build.
        """

        if project is None:
            project = self.project

        return os.path.abspath(
            os.path.join(os.environ[FACET_PATH_ENV], project, self.build_path_suffix)
        )

    def make_local_pypi_index_path(self) -> str:
        """
        Return the path where the local PyPi index for
        the given self.project should be placed.
        """
        return os.path.join(self.make_build_path(), "simple")

    def get_pyproject_toml(self) -> Dict[str, Any]:
        """
        Retrieve a parsed Dict for a given self.project's pyproject.toml.
        """
        pyproject_toml_path = os.path.join(
            os.environ[FACET_PATH_ENV], self.project, "pyproject.toml"
        )
        print(f"Reading build configuration from {pyproject_toml_path}")
        with open(pyproject_toml_path, "rt") as f:
            return toml.load(f)

    def get_package_dist_name(self) -> str:
        """
        Retrieves from pyproject.toml for a self.project the appropriate
        dist-name. E.g. "gamma-pytools" for self.project "pytools".
        """
        return self.get_pyproject_toml()[TOML_TOOL][TOML_FLIT][TOML_METADATA][
            TOML_DIST_NAME
        ]

    @abstractmethod
    def adapt_version_syntax(self, version: str) -> str:
        pass

    def expose_deps(self) -> None:
        """
        Expose package dependencies for builds as environment variables.
        """
        pyproject_toml = self.get_pyproject_toml()

        dependencies_to_expose = {}

        flit_metadata = pyproject_toml[TOML_TOOL][TOML_FLIT][TOML_METADATA]

        pkg_requires_python_version = flit_metadata[TOML_REQUIRES_PYTHON]

        dependencies_to_expose["python"] = pkg_requires_python_version

        default_deps_definition: List[str] = flit_metadata[TOML_REQUIRES]

        for default_dep in default_deps_definition:
            dependency_name: str
            dependency_version: str
            dependency_name, dependency_version = default_dep.strip().split(
                " ", maxsplit=1
            )
            dependencies_to_expose[dependency_name] = dependency_version.strip()

        if self.dependency_type != DEFAULT_DEPS:
            build_matrix_definition = pyproject_toml[TOML_BUILD][TOML_MATRIX]
            dependencies = build_matrix_definition[self.dependency_type]
            for (dependency_name, dependency_version) in dependencies.items():
                dependencies_to_expose[dependency_name] = self.adapt_version_syntax(
                    dependency_version
                )

        for d_name, d_version in dependencies_to_expose.items():
            # bash ENV variables can not use dash, replace it to _
            env_var_key = f"FACET_V_{d_name.replace('-','_')}".upper().strip()
            env_var_value = d_version.strip()
            print(f"Exposing: '{env_var_key}' as: '{env_var_value}'")
            os.environ[env_var_key] = env_var_value

    @abstractmethod
    def clean(self) -> None:
        """
        Cleans the dist folder for the given self.project and build system.
        """

    def print_build_info(self, stage: str) -> None:
        message = (
            f"{stage} {self.build_system.upper()} BUILD FOR {self.project}, "
            f"VERSION {self.package_version}"
        )
        separator = "=" * len(message)
        print(f"{separator}\n{message}\n{separator}")

    @abstractmethod
    def build(self) -> None:
        pass

    def create_local_pypi_index(self) -> None:
        """
        Creates/updates a local PyPI PEP 503 (the simple repository API) compliant
        folder structure, so that it can be used with PIP's --extra-index-url
        setting.
        """
        main_tox_build_path = self.make_build_path()
        pypi_index_path = self.make_local_pypi_index_path()
        project_dist_name = self.get_package_dist_name()
        project_repo_path = os.path.join(pypi_index_path, project_dist_name)
        project_index_html_path = os.path.join(project_repo_path, "index.html")
        os.makedirs(project_repo_path, exist_ok=True)

        package_glob = f"{project_dist_name}-*.tar.gz"

        # copy all relevant packages into the index subfolder
        for package in glob(os.path.join(main_tox_build_path, package_glob)):
            shutil.copy(package, project_repo_path)

        # remove index.html, if exists already
        if os.path.exists(project_index_html_path):
            os.remove(project_index_html_path)

        # create an index.html with entries for all existing packages
        package_file_links = [
            f"<a href='{os.path.basename(package)}'>{os.path.basename(package)}</a>"
            f"<br/>"
            for package in glob(os.path.join(project_repo_path, package_glob))
        ]
        # store index.html
        with open(project_index_html_path, "wt") as f:
            f.writelines(package_file_links)

        print(f"Local PyPi Index created at: {pypi_index_path}")

    def run(self) -> None:
        self.print_build_info(stage="STARTING")
        self.clean()
        self.expose_deps()
        self.build()
        self.print_build_info(stage="COMPLETED")


class CondaBuilder(Builder):
    @property
    def build_system(self) -> str:
        return B_CONDA

    @property
    def build_path_suffix(self) -> str:
        return CONDA_BUILD_PATH_SUFFIX

    def adapt_version_syntax(self, version: str) -> str:
        return version

    def clean(self) -> None:
        build_path = self.make_build_path()
        # purge pre-existing build directories
        package_dist_name = self.get_package_dist_name()
        for obsolete_folder in glob(os.path.join(build_path, f"{package_dist_name}_*")):
            print(f"Clean: Removing obsolete conda-build folder at: {obsolete_folder}")
            shutil.rmtree(obsolete_folder, ignore_errors=True)

        # remove broken packages
        shutil.rmtree(os.path.join(build_path, "broken"), ignore_errors=True)

    def build(self) -> None:
        """
        Build a facet self.project using conda-build.
        """

        build_path = self.make_build_path()
        os.environ[CONDA_BUILD_PATH_ENV] = build_path

        recipe_path = os.path.abspath(
            os.path.join(os.environ[FACET_PATH_ENV], self.project, "condabuild")
        )

        local_channels = [
            f'-c "{pathlib.Path(_build_path).as_uri()}"'
            for _build_path in map(self.make_build_path, KNOWN_PROJECTS)
            if os.path.exists(_build_path)
            and os.path.exists(os.path.join(_build_path, "noarch/repodata.json"))
        ]

        print(f"Building: {self.project}. Build path: {build_path}")
        os.makedirs(build_path, exist_ok=True)
        build_cmd = (
            f"conda-build -c conda-forge "
            f"-c bcg_gamma {' '.join(local_channels)} {recipe_path}"
        )
        print(f"Build Command: {build_cmd}")
        subprocess.run(args=build_cmd, shell=True, check=True)


class ToxBuilder(Builder):
    @property
    def build_system(self) -> str:
        return B_TOX

    @property
    def build_path_suffix(self) -> str:
        return TOX_BUILD_PATH_SUFFIX

    def adapt_version_syntax(self, version: str) -> str:
        # PIP expects == instead of =
        return re.sub(r"(?<![<=>~])=(?![<=>])", "==", version)

    def clean(self) -> None:
        # nothing to do – .tar.gz of same version will simply be replaced and
        # .tox is useful to keep
        pass

    def build(self) -> None:
        """
        Build a facet self.project using tox.
        """
        if self.dependency_type == DEFAULT_DEPS:
            tox_env = "py3"
        else:
            tox_env = "py3-custom-deps"

        build_cmd = f"tox -e {tox_env} -v"
        print(f"Build Command: {build_cmd}")
        subprocess.run(args=build_cmd, shell=True, check=True)
        print("Tox build completed – creating local PyPi index")
        self.create_local_pypi_index()


def print_usage() -> None:
    """
    Print a help string to explain the usage of this script.
    """
    usage = f"""Facet Build script
==================
Build a distribution package for given project.

Available arguments:
    project:    {' | '.join((ALL_PROJECTS_QUALIFIER,*KNOWN_PROJECTS))}
                use "{ALL_PROJECTS_QUALIFIER}" to build all projects

    build-system: {B_CONDA} | {B_TOX}

    dependencies:
        default: use dependencies and version ranges as defined in pyproject.toml
        min:     use a custom set of minimal dependencies from pyproject.toml
        max:     use a custom set of maximum dependencies from pyproject.toml


Example usage:
    ./make.py sklearndf conda default
    ./make.py sklearndf tox max

"""
    print(usage)


def run_make() -> None:
    """
    Run this build script with the given arguments.
    """
    if len(sys.argv) < 3:
        print_usage()
        exit(1)

    project = sys.argv[1]
    build_system = sys.argv[2]

    if len(sys.argv) > 3:
        dependency_type = sys.argv[3]
    else:
        dependency_type = DEFAULT_DEPS

    # sanitize input
    for arg_name, arg_passed, known_args in (
        ("project", project, KNOWN_PROJECTS),
        ("build system", build_system, KNOWN_BUILD_SYSTEMS),
        ("dependency type", dependency_type, KNOWN_DEPENDENCY_TYPES),
    ):

        if arg_name == "project" and arg_passed == "all":
            continue
        elif arg_passed not in known_args:
            print(
                f"Wrong {arg_name}: {arg_passed}; fallowed are: {', '.join(known_args)}"
            )
            exit(1)

    if project == ALL_PROJECTS_QUALIFIER:
        for _project in KNOWN_PROJECTS:
            Builder.for_build_system(
                build_system=build_system,
                project=_project,
                dependency_type=dependency_type,
            ).run()
    else:
        Builder.for_build_system(
            build_system=build_system, project=project, dependency_type=dependency_type
        ).run()


if __name__ == "__main__":
    run_make()



"""
Core implementation of :mod:`pytools.parallelization`.
"""
import logging
from typing import Any, Callable, Dict, Optional, Tuple, TypeVar, Union

import joblib

from ..api import AllTracker

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = ["ParallelizableMixin"]


#
# Type variables
#

T = TypeVar("T")


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


class ParallelizableMixin:
    """
    Mix-in class that supports parallelizing one or more operations using joblib.
    """

    def __init__(
        self,
        *,
        n_jobs: Optional[int] = None,
        shared_memory: Optional[bool] = None,
        pre_dispatch: Optional[Union[str, int]] = None,
        verbose: Optional[int] = None,
        **kwargs,
    ) -> None:
        """
        :param n_jobs: number of jobs to use in parallel;
            if ``None``, use joblib default (default: ``None``)
        :param shared_memory: if ``True``, use threads in the parallel runs; if
            ``False``, use multiprocessing (default: ``False``)
        :param pre_dispatch: number of batches to pre-dispatch;
            if ``None``, use joblib default (default: ``None``)
        :param verbose: verbosity level used in the parallel computation;
            if ``None``, use joblib default (default: ``None``)
        """
        super().__init__(**kwargs)
        #: Number of jobs to use in parallel; if ``None``, use joblib default.
        self.n_jobs = n_jobs

        #: If ``True``, use threads in the parallel runs;
        #: if ``False``, use multiprocessing.
        self.shared_memory = shared_memory

        #: Number of batches to pre-dispatch; if ``None``, use joblib default.
        self.pre_dispatch = pre_dispatch

        #: Verbosity level used in the parallel computation;
        #: if ``None``, use joblib default.
        self.verbose = verbose

        self._parallel_kwargs = {
            name: value
            for name, value in [
                ("n_jobs", n_jobs),
                ("require", "sharedmem" if shared_memory else None),
                ("pre_dispatch", pre_dispatch),
                ("verbose", verbose),
            ]
            if value is not None
        }

    def _parallel(self) -> joblib.Parallel:
        """
        Generate a :class:`joblib.Parallel` instance using the parallelization
        parameters of ``self``.

        :meta public:
        :return: the new :class:`joblib.Parallel` instance
        """
        return joblib.Parallel(**self._parallel_kwargs)

    @staticmethod
    def _delayed(
        function: Callable[..., T]
    ) -> Callable[..., Tuple[Callable[..., T], Tuple, Dict[str, Any]]]:
        """
        Decorate the given function for delayed execution.

        Convenience method preventing having to import :mod:`joblib`;
        defers to function :func:`joblib.delayed`.

        :meta public:
        :param function: the function to be delayed
        :return: the delayed function
        """
        return joblib.delayed(function)


__tracker.validate()



"""
Operators used in expressions.
"""
import logging
from abc import ABCMeta, abstractmethod
from typing import Mapping, Optional, Set, Tuple, TypeVar

from pytools.api import AllTracker

log = logging.getLogger(__name__)

__all__ = [
    "Operator",
    "BinaryOperator",
    "UnaryOperator",
]

__tracker = AllTracker(globals_=globals())


class Operator(metaclass=ABCMeta):
    """
    Base class for operators used in expressions.
    """

    #: Minimal value for :attr:`.precedence` (lower than the precedence of any
    #: Python operator).
    MIN_PRECEDENCE: int = -1

    #: Largest possible value of :attr:`.precedence` for default operators.
    MAX_PRECEDENCE: int

    def __init__(self, symbol: str, precedence: Optional[int] = None) -> None:
        """
        :param symbol: the symbol representing this operator
        :param precedence: the precedence of this operator; defaults to Python's
            standard precedence for symbols representing Python operators, and to
            :attr:`MIN_PRECEDENCE` for all other operators
        """
        self.symbol = symbol
        self._precedence = precedence

    @property
    @abstractmethod
    def is_unary(self) -> bool:
        """
        ``True`` if this is a unary operator.
        """
        pass

    @property
    def precedence(self) -> int:
        """
        The precedence of this operator (higher integers denote higher precedence).
        """
        return (
            _OPERATOR_PRECEDENCE.get(self, Operator.MIN_PRECEDENCE)
            if self._precedence is None
            else self._precedence
        )

    def __repr__(self) -> str:
        precedence = (
            f", precedence={self._precedence}" if self._precedence is not None else ""
        )
        return f'{type(self).__name__}("{self.symbol}"{precedence})'

    def __str__(self) -> str:
        return self.symbol


# Create a placeholder type so Sphinx will generate correct type hints
# for class attributes
# noinspection PyTypeHints,PyTypeChecker
_BinaryOperator = TypeVar("BinaryOperator", bound="BinaryOperator")


class BinaryOperator(Operator):
    """
    A binary operator.
    """

    #: Inherited from :attr:`.Operator.MIN_PRECEDENCE`.
    MIN_PRECEDENCE: int

    #: Inherited from :attr:`.Operator.MAX_PRECEDENCE`.
    MAX_PRECEDENCE: int

    DOT: _BinaryOperator  #: The ``.`` operator.
    POW: _BinaryOperator  #: The ``**`` operator.
    MUL: _BinaryOperator  #: The ``*`` operator.
    MATMUL: _BinaryOperator  #: The ``@`` operator.
    DIV: _BinaryOperator  #: The ``/`` operator (same as :attr:`.TRUE_DIV`).
    TRUE_DIV: _BinaryOperator  #: The ``/`` operator (same as :attr:`.DIV`).
    FLOOR_DIV: _BinaryOperator  #: The ``//`` operator.
    MOD: _BinaryOperator  #: The ``%`` operator.
    ADD: _BinaryOperator  #: The ``+`` operator.
    SUB: _BinaryOperator  #: The ``-`` operator.
    LSHIFT: _BinaryOperator  #: The ``<<`` operator.
    RSHIFT: _BinaryOperator  #: The ``>>`` operator.
    AND_BITWISE: _BinaryOperator  #: The ``&`` operator.
    XOR_BITWISE: _BinaryOperator  #: The ``^`` operator.
    OR_BITWISE: _BinaryOperator  #: The ``|`` operator.
    IN: _BinaryOperator  #: The ``in`` operator.
    NOT_IN: _BinaryOperator  #: The ``not in`` operator.
    IS: _BinaryOperator  #: The ``is`` operator.
    IS_NOT: _BinaryOperator  #: The ``is not`` operator.
    LT: _BinaryOperator  #: The ``<`` operator.
    LE: _BinaryOperator  #: The ``<=`` operator.
    GT: _BinaryOperator  #: The ``>`` operator.
    GE: _BinaryOperator  #: The ``>=`` operator.
    NEQ_: _BinaryOperator  #: The ``<>`` operator.
    NEQ: _BinaryOperator  #: The ``!=`` operator.
    EQ: _BinaryOperator  #: The ``==`` operator.
    AND: _BinaryOperator  #: The ``and`` operator.
    OR: _BinaryOperator  #: The ``or`` operator.
    ASSIGN: _BinaryOperator  #: The ``=`` operator.
    COLON: _BinaryOperator  #: The ``:`` operator.
    SLICE: _BinaryOperator  #: The ``:`` operator used in slice expressions.
    COMMA: _BinaryOperator  #: The ``,`` operator.
    NONE: _BinaryOperator  #: The empty operator.

    @property
    def is_unary(self) -> bool:
        """
        ``False``, as this is not a unary operator.
        """
        return False


# Create a placeholder type so Sphinx will generate correct type hints
# for class attributes
# noinspection PyTypeHints,PyTypeChecker
_UnaryOperator = TypeVar("UnaryOperator", bound="UnaryOperator")


class UnaryOperator(Operator):
    """
    A unary operator.
    """

    #: Inherited from :attr:`.Operator.MIN_PRECEDENCE`.
    MIN_PRECEDENCE: int

    #: Inherited from :attr:`.Operator.MAX_PRECEDENCE`.
    MAX_PRECEDENCE: int

    POS: _UnaryOperator  #: The ``+`` prefix operator.
    NEG: _UnaryOperator  #: The ``-`` prefix operator.
    INVERT: _UnaryOperator  #: The ``~`` prefix operator.
    NOT: _UnaryOperator  #: The ``not`` prefix operator.
    LAMBDA: _UnaryOperator  #: The ``lambda`` prefix operator.

    @property
    def is_unary(self) -> bool:
        """
        ``True``, as this is a unary operator.
        """
        return True


BinaryOperator.DOT = BinaryOperator(".")
BinaryOperator.POW = BinaryOperator("**")
BinaryOperator.MUL = BinaryOperator("*")
BinaryOperator.MATMUL = BinaryOperator("@")
BinaryOperator.DIV = BinaryOperator.TRUE_DIV = BinaryOperator("/")
BinaryOperator.FLOOR_DIV = BinaryOperator("//")
BinaryOperator.MOD = BinaryOperator("%")
BinaryOperator.ADD = BinaryOperator("+")
BinaryOperator.SUB = BinaryOperator("-")
BinaryOperator.LSHIFT = BinaryOperator("<<")
BinaryOperator.RSHIFT = BinaryOperator(">>")
BinaryOperator.AND_BITWISE = BinaryOperator("&")
BinaryOperator.XOR_BITWISE = BinaryOperator("^")
BinaryOperator.OR_BITWISE = BinaryOperator("|")
BinaryOperator.IN = BinaryOperator("in")
BinaryOperator.NOT_IN = BinaryOperator("not in")
BinaryOperator.IS = BinaryOperator("is")
BinaryOperator.IS_NOT = BinaryOperator("is not")
BinaryOperator.LT = BinaryOperator("<")
BinaryOperator.LE = BinaryOperator("<=")
BinaryOperator.GT = BinaryOperator(">")
BinaryOperator.GE = BinaryOperator(">=")
BinaryOperator.NEQ_ = BinaryOperator("<>")
BinaryOperator.NEQ = BinaryOperator("!=")
BinaryOperator.EQ = BinaryOperator("==")
BinaryOperator.AND = BinaryOperator("and")
BinaryOperator.OR = BinaryOperator("or")
BinaryOperator.ASSIGN = BinaryOperator("=")
BinaryOperator.COLON = BinaryOperator(":")
BinaryOperator.SLICE = BinaryOperator(":")
BinaryOperator.COMMA = BinaryOperator(",")
BinaryOperator.NONE = BinaryOperator("")

UnaryOperator.POS = UnaryOperator("+")
UnaryOperator.NEG = UnaryOperator("-")
UnaryOperator.INVERT = UnaryOperator("~")
UnaryOperator.NOT = UnaryOperator("not")
UnaryOperator.LAMBDA = UnaryOperator("lambda")

__OPERATOR_PRECEDENCE_ORDER: Tuple[Set[Operator], ...] = (
    {BinaryOperator.ASSIGN},
    {BinaryOperator.COMMA, UnaryOperator.LAMBDA},
    {BinaryOperator.COLON, BinaryOperator.SLICE},
    {BinaryOperator.OR},
    {BinaryOperator.AND},
    {UnaryOperator.NOT},
    {BinaryOperator.NEQ_, BinaryOperator.NEQ, BinaryOperator.EQ},
    {
        BinaryOperator.IN,
        BinaryOperator.NOT_IN,
        BinaryOperator.IS,
        BinaryOperator.IS_NOT,
        BinaryOperator.LT,
        BinaryOperator.LE,
        BinaryOperator.GT,
        BinaryOperator.GE,
    },
    {BinaryOperator.OR_BITWISE},
    {BinaryOperator.XOR_BITWISE},
    {BinaryOperator.AND_BITWISE},
    {BinaryOperator.LSHIFT, BinaryOperator.RSHIFT},
    {BinaryOperator.ADD, BinaryOperator.SUB},
    {
        BinaryOperator.MUL,
        BinaryOperator.MATMUL,
        BinaryOperator.TRUE_DIV,
        BinaryOperator.FLOOR_DIV,
        BinaryOperator.MOD,
    },
    {UnaryOperator.POS, UnaryOperator.NEG},
    {UnaryOperator.INVERT},
    {BinaryOperator.POW},
    {BinaryOperator.DOT},
)

_OPERATOR_PRECEDENCE: Mapping[Operator, int] = {
    operator: priority
    for priority, operators in enumerate(__OPERATOR_PRECEDENCE_ORDER)
    for operator in operators
}

Operator.MAX_PRECEDENCE = len(__OPERATOR_PRECEDENCE_ORDER)

__tracker.validate()



"""
Supporting classes for linkage trees.
"""

import logging
from abc import ABCMeta, abstractmethod
from typing import Optional

from pytools.api import AllTracker, inheritdoc

log = logging.getLogger(__name__)

#
# exported names
#

__all__ = ["Node", "LeafNode", "LinkageNode"]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


class Node(metaclass=ABCMeta):
    """
    Base class for nodes of a :class:`.LinkageTree`.
    """

    def __init__(self, index: int) -> None:
        """
        :param index: the numerical index of this node in the linkage tree
        """
        self._index = index

    @property
    def index(self) -> int:
        """
        The numerical index of this node in the linkage tree.
        """
        return self._index

    @property
    @abstractmethod
    def children_distance(self) -> float:
        """
        Distance of this node from its children.
        """
        pass

    @property
    @abstractmethod
    def weight(self) -> float:
        """
        The weight of this node.
        """
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """
        The name of this node.
        """
        pass

    @property
    @abstractmethod
    def is_leaf(self) -> bool:
        """
        ``True`` if the node is a leaf; ``False`` otherwise.
        """
        pass

    def _type_error(self, property_name: str) -> TypeError:
        return TypeError(f"{property_name} is not defined for a {type(self).__name__}")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}_{self._index}"


@inheritdoc(match="[see superclass]")
class LinkageNode(Node):
    """
    An inner node in a :class:`.LinkageTree`.
    """

    def __init__(self, index: int, children_distance: Optional[float]) -> None:
        """
        :param children_distance: the distance between this node and its children
        """
        super().__init__(index=index)
        self._children_distance = children_distance

    __init__.__doc__ = Node.__init__.__doc__ + __init__.__doc__

    @property
    def children_distance(self) -> float:
        """[see superclass]"""
        return self._children_distance

    @property
    def weight(self) -> float:
        """
        Undefined for inner nodes; raises :class:`TypeError`.
        """
        raise self._type_error("weight")

    @property
    def name(self) -> str:
        """
        Undefined for inner nodes; raises :class:`TypeError`.
        """
        raise self._type_error("name")

    @property
    def is_leaf(self) -> bool:
        """``False``"""
        return False

    def __repr__(self) -> str:
        return f"{super().__repr__()}[dist={self.children_distance * 100:.0f}%]"


@inheritdoc(match="[see superclass]")
class LeafNode(Node):
    """
    A leaf in a :class:`.LinkageTree`.
    """

    def __init__(self, index: int, name: str, weight: float) -> None:
        """
        :param name: the name of the leaf
        :param weight: the weight of the leaf
        """
        super().__init__(index=index)
        self._name = name
        self._weight = weight

    __init__.__doc__ = Node.__init__.__doc__ + __init__.__doc__

    @property
    def children_distance(self) -> float:
        """
        Undefined for leaf nodes; raises :class:`TypeError`.
        """
        raise self._type_error("children_distance")

    @property
    def name(self) -> str:
        """[see superclass]"""
        return self._name

    @property
    def weight(self) -> float:
        """[see superclass]"""
        return self._weight

    @property
    def is_leaf(self) -> bool:
        """``True``"""
        return True

    def __repr__(self) -> str:
        return f"{super().__repr__()}[name={self.name}, weight={self.weight}]"


__tracker.validate()



"""
Matplot styles for the GAMMA visualization library.
"""

import logging
from abc import ABCMeta
from typing import Optional, Tuple

import matplotlib.pyplot as plt
from matplotlib import text as mt
from matplotlib.axes import Axes
from matplotlib.backend_bases import RendererBase
from matplotlib.colorbar import ColorbarBase, make_axes
from matplotlib.colors import Normalize
from matplotlib.legend import Legend
from matplotlib.ticker import Formatter
from matplotlib.tight_layout import get_renderer

from ..api import AllTracker, inheritdoc
from ._viz import ColoredStyle
from .color import MatplotColorScheme, RgbaColor

log = logging.getLogger(__name__)

#
# Exported names
#

__all__ = ["MatplotStyle", "ColorbarMatplotStyle"]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


@inheritdoc(match="[see superclass]")
class MatplotStyle(ColoredStyle[MatplotColorScheme], metaclass=ABCMeta):
    """
    Base class of drawing styles using `matplotlib`.
    """

    def __init__(
        self,
        *,
        ax: Optional[Axes] = None,
        colors: Optional[MatplotColorScheme] = None,
        **kwargs,
    ) -> None:
        """
        :param ax: optional axes object to draw on; create a new figure if not specified
        """
        super().__init__(colors=colors, **kwargs)
        self._ax = ax
        self._renderer: Optional[RendererBase] = None

    __init__.__doc__ += ColoredStyle.__init__.__doc__

    @classmethod
    def get_default_style_name(cls) -> str:
        """[see superclass]"""
        return "matplot"

    @property
    def ax(self) -> Axes:
        """
        The matplotlib :class:`~matplotlib.axes.Axes` object on which to draw.
        """
        ax = self._ax
        if ax is None:
            _, ax = plt.subplots()
            self._ax = ax
        return ax

    @property
    def renderer(self) -> RendererBase:
        """
        The renderer used by this style's :class:`~matplotlib.axes.Axes` object
        (see :attr:`.ax`).
        """
        renderer = self._renderer
        if renderer is None:
            self._renderer = renderer = get_renderer(self.ax.figure)
        return renderer

    def text_dimensions(
        self, text: str, x: Optional[float] = None, y: Optional[float] = None, **kwargs
    ) -> Tuple[float, float]:
        """
        Calculate the horizontal and vertical dimensions of the given text in axis
        units.

        Constructs a :class:`~matplotlib.text.Text` artist then calculates it size
        relative to the axis :attr:`.ax` managed by this style object.
        For non-linear axis scales, text size differs depending on placement,
        so the intended placement (in data coordinates) should be provided.

        :param text: text to calculate the dimensions for
        :param x: intended horizontal text placement (optional, defaults to left of
            view)
        :param y: intended vertical text placement (optional, defaults to bottom of
            view)
        :param kwargs: additional keyword arguments to use when constructing the
            :class:`~matplotlib.text.Text` artist, e.g., rotation
        :return: tuple `(width, height)` in axis units
        """

        ax = self.ax

        if x is None or y is None:
            x0, y0, _, _ = ax.dataLim.bounds
            if x is None:
                x = x0
            if y is None:
                y = y0

        fig = ax.figure

        extent = mt.Text(x, y, text, figure=fig, **kwargs).get_window_extent(
            self.renderer
        )

        (x0, y0), (x1, y1) = ax.transData.inverted().transform(extent)

        return abs(x1 - x0), abs(y1 - y0)

    def start_drawing(self, title: str, **kwargs) -> None:
        """
        Set the title of the matplot chart to the given title, and set the foreground
        and background color according to the color scheme.

        :param title: the chart title
        """

        ax = self.ax

        # set title plus title color
        ax.set_title(label=title, color=self.colors.foreground)

        # color the axes
        self._apply_color_scheme(ax)

        bg_color = self.colors.background

        # set the figure background color
        try:
            # does the axes background color conflict with the figure background color?
            if ax.figure.__pytools_viz_background != bg_color:
                log.warning(
                    "subplots have conflicting color schemes; setting background color "
                    f'according to color scheme {self.colors!r} of chart "{title}"'
                )
        except AttributeError:
            pass
        ax.figure.set_facecolor(bg_color)
        ax.figure.__pytools_viz_background = bg_color

    def finalize_drawing(self, **kwargs) -> None:
        """[see superclass]"""

        super().finalize_drawing(**kwargs)

        # set legend color
        legend: Legend = self.ax.get_legend()

        if legend:
            patch = legend.legendPatch
            fg_color = self.colors.foreground

            patch.set_facecolor(self.colors.background)
            patch.set_alpha(0.5)
            patch.set_edgecolor(fg_color)

            for text in legend.get_texts():
                text.set_color(fg_color)

    def _apply_color_scheme(self, ax: Axes) -> None:
        """
        Apply this style's color scheme to the given :class:`~matplotlib.axes.Axes`.

        Style implementations can use this to apply the color scheme to sub-axes.
        Does not need to be applied to main axes, as this is already done in method
        :meth:`.start_drawing`.

        This method will be public as of v1.1.

        :param ax: the axes to apply the color scheme to
        """

        fg_color = self.colors.foreground
        bg_color = self.colors.background

        # set face color
        ax.set_facecolor(bg_color)

        # set figure face color
        ax.patch.set_facecolor(bg_color)

        # set tick and tick label color
        ax.tick_params(color=fg_color, labelcolor=fg_color)

        # set the outline color
        for spine in ax.spines.values():
            spine.set_edgecolor(fg_color)


class ColorbarMatplotStyle(MatplotStyle, metaclass=ABCMeta):
    """
    `matplotlib` style with added support for a color bar.

    The plot uses a color map to indicate a scalar value,
    and the color bar provides a legend for the color gradient.

    The color map is provided by the :class:`.MatplotColorScheme` associated with the
    style.
    """

    def __init__(
        self,
        *,
        ax: Optional[Axes] = None,
        colors: Optional[MatplotColorScheme] = None,
        colormap_normalize: Normalize = None,
        colorbar_major_formatter: Optional[Formatter] = None,
        colorbar_minor_formatter: Optional[Formatter] = None,
        **kwargs,
    ):
        """
        :param colormap_normalize: the :class:`~matplotlib.colors.Normalize` object
            that maps values to color indices; unless otherwise specified, use a plain
            :class:`~matplotlib.colors.Normalize` object with linear autoscaling
        :param colorbar_major_formatter: major tick formatter for the color bar
            (optional)
        :param colorbar_minor_formatter: minor tick formatter for the color bar
            (optional; requires that also a major formatter is specified)
        """
        super().__init__(ax=ax, colors=colors, **kwargs)

        self.colormap_normalize = (
            Normalize() if colormap_normalize is None else colormap_normalize
        )
        if colorbar_minor_formatter is not None and colorbar_major_formatter is None:
            raise ValueError(
                "arg colorbar_minor_formatter passed without passing "
                "arg colorbar_major_formatter"
            )
        self.colorbar_major_formatter = colorbar_major_formatter
        self.colorbar_minor_formatter = colorbar_minor_formatter

        self.colorbar = None

    __init__.__doc__ = MatplotStyle.__init__.__doc__ + __init__.__doc__

    def color_for_value(self, z: float) -> RgbaColor:
        """
        Get the color associated with a given scalar, based on the color map and
        normalization defined for this style.

        :param z: the scalar to be color-encoded
        :return: the resulting color as a RGBA tuple
        """
        return self.colors.colormap(self.colormap_normalize(z))

    def finalize_drawing(self, colorbar_label: Optional[str] = None, **kwargs) -> None:
        """
        Add the color bar to the chart.

        :param colorbar_label: the label for the color bar
        """

        super().finalize_drawing(**kwargs)

        fg_color = self.colors.foreground

        cax, _ = make_axes(self.ax)
        self.colorbar = cb = ColorbarBase(
            cax,
            cmap=self.colors.colormap,
            norm=self.colormap_normalize,
            label=colorbar_label or "",
            orientation="vertical",
        )

        if self.colorbar_major_formatter is not None:
            cax.yaxis.set_major_formatter(self.colorbar_major_formatter)

        if self.colorbar_minor_formatter is not None:
            cax.yaxis.set_minor_formatter(self.colorbar_minor_formatter)

        # set colorbar tick color
        cb.ax.yaxis.set_tick_params(colors=fg_color)

        # set colorbar edge color
        cb.outline.set_edgecolor(fg_color)


__tracker.validate()



"""
Representations of Python expressions and support for pretty-printed multi-line output.

Useful for generating representations of complex Python objects.

The simplest expressions are `atomic` identifiers and literals, represented by
instances of :class:`.Id` and :class:`.Lit`.
To create an identifier, use one of

.. code-block:: python

    Id.x
    Id("x")

To create identifiers with a leading or trailing underscore, you need to use the
:class:`.Id` constructor as demonstrated in the second variant above.

.. code-block:: python

    Id._x     # will raise exception
    Id.x_     # will raise exception
    Id("_x")  # works
    Id("x_")  # works

To create a literal, use one of

.. code-block:: python

    Lit(3)
    Lit(3.3)
    Lit("text")

See :func:`.make_expression` below for more ways to create literals.

You can use most Python operators to create composite expressions – even using literals
without :class:`.Lit` where the context makes it clear that an expression object is
being constructed, e.g.:

.. code-block:: python

    x, y, z, f, g = Id.x, Id.y, Id.z, Id.f, Id.g
    x + 3   # x is an Expression instance, hence 3 is converted to a Literal object
    3 + x   # this also works if only the second argument is an Expression object
    x == y
    f(3, 5, z)
    g[5]
    g.a

Function :func:`.make_expression` translates most Python objects to useful expressions:

.. code-block:: python

    make_expression(3)
    make_expression(3.3)
    make_expression("text")
    make_expression([3, 5, Id.x])
    make_expression({"a": 3, "b": 5, "c": 6})

If you pass named object to :func:`.make_expression`, e.g., a function or a class,
it will return an identifier with the name of that object.

.. code-block:: python

    make_expression(my_function)
    make_expression(MyClass)

Note that ``a == b`` does not compare two expressions for equality, but creates a
comparison expression.
To compare two expressions, instead use ``a.eq_(b)``, or convert both expressions to
`frozen expressions` using :func:`.freeze`, thus removing their ability to create new
expressions using Python operators: ``freeze(a) == freeze(b)``.

Also note the trailing underscore of method :meth:`~.Expression.eq_`.
By convention, all methods and attributes of expression objects have a trailing
underscore, to distinguish them from the shortcut to generate attribute access
expressions: ``a.eq(b)`` creates a new expression object, rather than comparing
two expressions.

Finally, :class:`.HasExpressionRepr` is a mix-in class that provides a simple way
for any class to produce well-formatted :func:`.repr` string representations
by implementing method :meth:`~.HasExpressionRepr.to_expression` and generating
an :class:`Expression` representation of themselves.

Even class :class:`Expression` subclasses :class:`.HasExpressionRepr`;
:meth:`.Expression.to_expression` returns the expression itself.
"""

from ._expression import *



"""
Test the pytools.viz module
"""


def test_imports() -> None:
    from pytools.viz import color, dendrogram, distribution, matrix, util
    from pytools.viz.dendrogram import base as dendrogram_base

    # we need this assertion to prevent an "unused imports" error during linting
    assert all(
        module is not None
        for module in [
            color,
            dendrogram,
            dendrogram_base,
            distribution,
            matrix,
            util,
        ]
    )



"""
Core implementation of :mod:`pytools.api`.
"""

import inspect
import logging
import re
import warnings
from functools import wraps
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import numpy as np
import pandas as pd
import typing_inspect

log = logging.getLogger(__name__)

__all__ = [
    "AllTracker",
    "is_list_like",
    "to_tuple",
    "to_list",
    "to_set",
    "validate_type",
    "validate_element_types",
    "get_generic_bases",
    "public_module_prefix",
    "deprecated",
    "deprecation_warning",
    "inheritdoc",
]

T = TypeVar("T")
T_Collection = TypeVar("T_Collection", bound=Collection)


class AllTracker:
    """
    Track global, public items defined in a module and validate that all eligible items
    have been included in the ``__all__`` variable.

    Items will be tracked if:

    - their name does not start with an underscore character
    - they are defined after the :class:`.AllTracker` instance has been created

    The underlying idea (and widely used pattern in GAMMA packages) is to define all
    items in one or more private submodules, export all public items using the
    ``__all__`` keyword, and import them into a public package one level up
    (usually the package's ``__init__.py``).

    This ensures a clean namespace in the public package, uncluttered by any imports
    used in the private module.
    """

    def __init__(self, globals_: Dict[str, Any], public_module: Optional[str] = None):
        """
        :param globals_: the dictionary of global variables returned by calling
            :meth:`._globals` in the current module scope
        :param public_module: full name of the public module that will export the items
            managed by this tracker

        """
        self._globals = globals_
        self._imported = set(globals_.keys())

        if public_module:
            self.public_module = public_module

        else:
            try:
                module = globals_["__name__"]
            except KeyError:
                raise ValueError(
                    "cannot infer public module: "
                    "arg globals_ does not define module name in __name__"
                )

            self.public_module = public_module_prefix(module)

    #: Full name of the public module that will export the items managed by this
    #: tracker.
    public_module: str

    def validate(self) -> None:
        """
        Validate that all eligible items that were defined since the creation of this
        tracker are listed in the ``__all__`` variable.

        :raise RuntimeError: if ``__all__`` is not as expected
        """
        all_expected = self.get_tracked()

        globals_ = self._globals

        if set(globals_.get("__all__", [])) != set(all_expected):
            raise RuntimeError(
                "missing or unexpected all declaration, "
                f"expected:\n__all__ = {all_expected}"
            )

        public_module = self.public_module

        for name in all_expected:
            try:
                globals_[name].__publicmodule__ = public_module
            except AttributeError:
                pass

    def get_tracked(self) -> List[str]:
        """
        List the names of all locally tracked, public items.

        :return: the list of object names, sorted alphabetically
        """
        return sorted(name for name in self._globals if self._is_eligible(name=name))

    def is_tracked(self, name: str, item: Any) -> bool:
        """
        Check if the given item is tracked by this tracker under the given name.

        :param name: the name of the item in the tracker's global namespace
        :param item: the item referred to by the name
        """
        try:
            return self._globals[name] is item and self._is_eligible(name)
        except KeyError:
            return False

    def _is_eligible(self, name: str) -> bool:
        # check if the given item is eligible for tracking by this tracker
        return not (name.startswith("_") or name in self._imported)

    def __getitem__(self, name: str) -> Any:
        # get a tracked item by name

        item = self._globals[name]
        if self._is_eligible(name=name):
            return item
        else:
            raise KeyError(name)


# regular expression to extract the public prefix of a private module
# this is the module path up to the first submodule with a leading underscore
__RE_PUBLIC_MODULE = re.compile(
    # start of match group for the public part of the module path
    r"("
    # first module component must not start with '_')
    r"(?:[a-zA-Z]\w+)"
    # then as many path components as can be matched non-greedily …)
    r"(?:\.\w+)*?"
    # but when we hit the first private path component, we stop matching the
    # public part of the path …
    r")"
    # … and match the rest as the private path
    r"(?:\._\w*(?:\.\w+)*)?"
)


def public_module_prefix(module_name: str) -> str:
    """
    Get the public prefix of the given module name.

    A module name is composed of one or more identifiers, separated by ``.`` operators.
    The public prefix is the module name including all identifiers up to, and excluding,
    the first identifier starting with an underscore character (``_``) .
    If there is no such identifier, then the public prefix is the same as the full
    module name.
    If the first identifier starts with a ``_``, then the public prefix is not defined.

    For example:

    - the public module prefix of ``a.b._c.d._e`` is ``a.b``
    - the public module prefix of ``a.b`` is ``a.b``
    - the public module prefix of ``_a.b.c`` is undefined

    :param module_name: the module name for which to get the public prefix
    :return: the public prefix
    :raise ValueError: the public prefix is undefined
    """
    match = __RE_PUBLIC_MODULE.fullmatch(module_name)
    if not match:
        raise ValueError(f"cannot infer public module path from module {module_name}")
    return match[1]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())
# we forget that class AllTracker itself is already defined, because we want to include
# it in the __all__ statement
# noinspection PyProtectedMember
__tracker._imported.remove(AllTracker.__name__)
# noinspection PyProtectedMember
__tracker._imported.remove(public_module_prefix.__name__)

#
# Functions
#


def is_list_like(obj: Any) -> bool:
    """
    Check if the object is list-like.

    Objects that are considered list-like when they implement methods ``len`` and
    ``__getitem__``. These include, for example, lists, tuples, sets, NumPy arrays, and
    Pandas series and indices.

    As an exception, the following types are not considered list-like:

    - :class:`str`
    - :class:`bytes`
    - :class:`pandas.DataFrame`: inconsistent behaviour of the sequence interface;
        iterating a data frame yields the values of the column index, while the length
        of a data frame is its number of rows
    - :class:`numpy.ndarray` instances with 0 dimensions

    :param obj: The object to check
    :return: ``True`` if ``obj`` has list-like properties, ``False`` otherwise
    """

    return (
        hasattr(obj, "__len__")
        and hasattr(obj, "__getitem__")
        and not isinstance(obj, (str, bytes))
        # pandas data objects with more than 1 dimension, e.g., data frames
        and not (isinstance(obj, pd.NDFrame) and obj.ndim != 1)
        # exclude zero-dimensional numpy arrays, effectively scalars
        and not (isinstance(obj, np.ndarray) and obj.ndim == 0)
    )


def to_tuple(
    values: Union[Iterable[T], T],
    *,
    element_type: Optional[Type[T]] = None,
    arg_name: Optional[str] = None,
) -> Tuple[T, ...]:
    """
    Return the given values as a tuple.

    - if arg `values` is a tuple, return arg `values` unchanged
    - if arg `values` is an iterable other than a tuple, return a list of its elements
    - if arg `values` is not an iterable, return a tuple with the value as its only
      element

    :param values: one or more elements to return as a tuple
    :param element_type: expected type of the values; raise a ``TypeException`` if one
        or more values do not implement this type
    :param arg_name: name of the argument as which the values were passed to a function
        or method; used when composing the ``TypeException`` message
    :return: the values as a tuple
    :raise TypeException: one or more values did not match the expected type
    """

    return _to_collection(
        values=values,
        collection_type=tuple,
        element_type=element_type,
        arg_name=arg_name,
    )


def to_list(
    values: Union[Iterable[T], T],
    *,
    element_type: Optional[Type[T]] = None,
    arg_name: Optional[str] = None,
) -> List[T]:
    """
    Return the given values as a list.

    - if arg `values` is a list, return arg `values` unchanged
    - if arg `values` is an iterable other than a list, return a list of its elements
    - if arg `values` is not an iterable, return a list with the value as its only
      element

    :param values: one or more elements to return as a list
    :param element_type: expected type of the values; raise a ``TypeException`` if one
        or more values do not implement this type
    :param arg_name: name of the argument as which the values were passed to a function
        or method; used when composing the ``TypeException`` message
    :return: the values as a list
    :raise TypeException: one or more values did not match the expected type
    """

    return _to_collection(
        values=values,
        collection_type=list,
        element_type=element_type,
        arg_name=arg_name,
    )


def to_set(
    values: Union[Iterable[T], T],
    *,
    element_type: Optional[Type[T]] = None,
    arg_name: Optional[str] = None,
) -> Set[T]:
    """
    Return the given values as a set.

    - if arg `values` is a set, return arg `values` unchanged
    - if arg `values` is an iterable other than a set, return a set of its elements
    - if arg `values` is not an iterable, return a list with the value as its only
      element

    :param values: one or more elements to return as a set
    :param element_type: expected type of the values; raise a ``TypeException`` if one
        or more values do not implement this type
    :param arg_name: name of the argument as which the values were passed to a function
        or method; used when composing the ``TypeException`` message
    :return: the values as a set
    :raise TypeException: one or more values did not match the expected type
    """

    return _to_collection(
        values=values, collection_type=set, element_type=element_type, arg_name=arg_name
    )


def _to_collection(
    values: Union[T, Iterable[T]],
    *,
    collection_type: Type[T_Collection],
    element_type: Optional[Type[T]],
    arg_name: Optional[str] = None,
) -> T_Collection:

    elements: T_Collection

    if (
        isinstance(values, Iterable)
        and not isinstance(values, str)
        and not isinstance(values, bytes)
    ):
        if isinstance(values, collection_type):
            # no change needed, values already is the collection we need
            elements = values
        elif element_type and isinstance(values, element_type):
            # create a single-element collection
            elements = collection_type((values,))
        else:
            elements = collection_type(values)
    else:
        # create a single-element collection
        elements = collection_type((values,))

    if element_type:
        validate_element_types(
            elements, expected_type=element_type, name=f"arg {arg_name}"
        )

    return elements


def validate_type(
    value: T,
    *,
    expected_type: Type[T],
    optional: bool = False,
    name: Optional[str] = None,
) -> None:
    """
    Validate that a value implements the expected type.

    :param value: an arbitrary object
    :param expected_type: the type to check for
    :param optional: if ``True``, accept ``None`` as a valid value (default: ``False``)
    :param name: optional name of the argument or callable with/to which the value
        was passed; use ``"arg …"`` for arguments, or the name of a callable if
        verifying positional arguments
    :raise TypeException: the value did not match the expected type
    """
    if expected_type == object:
        return

    if optional and value is None:
        return

    if not isinstance(value, expected_type):
        if name:
            message_head = f"{name} requires"
        else:
            message_head = "expected"
        raise TypeError(
            f"{message_head} instance of {expected_type.__name__} "
            f"but got a {type(value).__name__}"
        )


def validate_element_types(
    iterable: Iterable[T], *, expected_type: Type[T], name: Optional[str] = None
) -> None:
    """
    Validate that all elements in the given iterable implement the expected type.

    :param iterable: an iterable
    :param expected_type: the type to check for
    :param name: optional name of the argument or callable with/to which the elements
        were passed; use ``"arg …"`` for arguments, or the name of a callable if
        verifying positional arguments
    :raise TypeException: one or more elements of the iterable did not match the
        expected type
    """
    if expected_type == object:
        return

    for element in iterable:
        if not isinstance(element, expected_type):
            if name:
                message_head = f"{name} requires"
            else:
                message_head = "expected"
            raise TypeError(
                f"{message_head} instances of {expected_type.__name__} "
                f"but got a {type(element).__name__}"
            )


def get_generic_bases(cls: type) -> Tuple[type, ...]:
    """
    Bugfix version of :func:`typing_inspect.get_generic_bases`.

    Prevents getting the generic bases of the parent class if not defined for the given
    class.

    :param cls: class to get the generic bases for
    :return: the generic base classes of the given class
    """
    bases = typing_inspect.get_generic_bases(cls)
    if bases is typing_inspect.get_generic_bases(super(cls, cls)):
        return ()
    else:
        return bases


#
# Decorators
#


def deprecated(function: Callable = None, *, message: Optional[str] = None):
    """
    Decorator to mark a function as deprecated.

    Logs a warning when the decorated function is called.

    To deprecate classes, apply this decorator to the ``__init__`` method, not to the
    class itself.

    :param function: the function to be decorated
    :param message: custom message to include when logging the warning (optional)
    """

    def _deprecated_inner(func: callable) -> callable:
        @wraps(func)
        def new_func(*args, **kwargs) -> Any:
            """
            Function wrapper
            """
            message_header = (
                f"Call to deprecated {type(func).__name__} {func.__qualname__}"
            )
            if message is None:
                warnings.warn(message_header, FutureWarning, stacklevel=2)
            else:
                warnings.warn(
                    f"{message_header}: {message}", FutureWarning, stacklevel=2
                )
            return func(*args, **kwargs)

        return new_func

    if function is None:
        return _deprecated_inner
    elif callable(function):
        return _deprecated_inner(function)
    elif isinstance(function, str):
        raise ValueError(
            "Deprecation message not provided as a keyword argument. "
            f'Usage: @{deprecated.__name__}(message="...")'
        )
    else:
        raise ValueError("Deprecated object must be callable")


def deprecation_warning(message: str, stacklevel: int = 1) -> None:
    """
    Issue a deprecation warning.

    :param message: the warning message
    :param stacklevel: stack level relative to caller for emitting the context of the
        warning (default: 1)
    :return:
    """
    if stacklevel < 1:
        raise ValueError(f"arg stacklevel={stacklevel} must be a positive integer")
    warnings.warn(message, FutureWarning, stacklevel=stacklevel + 1)


# noinspection PyIncorrectDocstring
def inheritdoc(cls: type = None, *, match: str) -> Union[type, Callable[[type], type]]:
    """
    Decorator to inherit docstrings of overridden methods.

    Usage:

    .. code-block:: python

      class A:
          def my_function(self) -> None:
          \"""Some documentation\"""
          # …

      @inheritdoc(match="[see superclass]")
      class B(A):
          def my_function(self) -> None:
          \"""[see superclass]\"""
          # …

          def my_other_function(self) -> None:
          \"""This docstring will not be replaced\"""
          # …

    In this example, the docstring of the ``my_function`` will be replaced with the
    docstring of the overridden function of the same name, or with ``None`` if no
    overridden function exists, or if that function has no docstring.

    :param match: the parent docstring will be inherited if the current docstring
        is equal to match
    """

    def _inheritdoc_inner(_cls: type) -> type:
        if not type(_cls):
            _raise_type_error(_cls)

        match_found = False

        def _get_docstring(m: Any) -> str:
            try:
                return m.__func__.__doc__
            except AttributeError:
                return m.__doc__

        def _set_docstring(m: Any, d: str) -> None:
            try:
                m.__func__.__doc__ = d
            except AttributeError:
                m.__doc__ = d

        for name, member in vars(_cls).items():
            doc = _get_docstring(m=member)
            if doc == match:
                parents = inspect.getmro(_cls)[1:]
                _set_docstring(
                    m=member,
                    d=(
                        _get_docstring(m=getattr(parents[0], name, None))
                        if parents
                        else None
                    ),
                )
                match_found = True

        if not match_found:
            log.warning(
                f"{inheritdoc.__name__}:"
                f"no match found for docstring {repr(match)} in class {_cls.__name__}"
            )

        return _cls

    def _raise_type_error(_cls: type) -> None:
        raise TypeError(
            f"@{inheritdoc.__name__} can only decorate classes, "
            f"not a {type(_cls).__name__}"
        )

    if cls is None:
        return _inheritdoc_inner
    elif type(cls):
        return _inheritdoc_inner(cls)
    elif isinstance(cls, str):
        raise ValueError(
            "arg match not provided as a keyword argument. "
            f'Usage: @{inheritdoc.__name__}(match="...")'
        )
    else:
        _raise_type_error(cls)


__tracker.validate()



"""
Base classes for matrix styles.
"""

import logging
from abc import ABCMeta, abstractmethod

import pandas as pd

from pytools.api import AllTracker
from pytools.viz import DrawingStyle

log = logging.getLogger(__name__)

#
# Exported names
#

__all__ = ["MatrixStyle"]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


class MatrixStyle(DrawingStyle, metaclass=ABCMeta):
    """
    Base class for matrix drawer styles.
    """

    @abstractmethod
    def draw_matrix(self, matrix: pd.DataFrame) -> None:
        """
        Draw the matrix.

        :param matrix: the matrix represented as a data frame
        """
        pass


__tracker.validate()



"""
Text styles for the GAMMA visualization library.
"""

import logging
import sys
from abc import ABCMeta
from typing import Optional, TextIO

from ..api import AllTracker, inheritdoc
from ._viz import DrawingStyle

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = ["TextStyle"]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


@inheritdoc(match="[see superclass]")
class TextStyle(DrawingStyle, metaclass=ABCMeta):
    """
    Base class of drawing styles producing plain text output.
    """

    #: The output stream this style instance writes to.
    out: TextIO

    #: The maximum width of the text to be produced.
    width: int

    def __init__(self, out: Optional[TextIO] = None, width: int = 80, **kwargs) -> None:
        """
        :param width: the maximum width available to render the text, defaults to 80
        :param out: the output stream this style instance writes to
            (defaults to :obj:`sys.stdout`)
        """

        super().__init__(**kwargs)

        if width <= 0:
            raise ValueError(
                f"arg width expected to be positive integer but is {width}"
            )
        self.out = sys.stdout if out is None else out
        self.width = width

    @classmethod
    def get_default_style_name(cls) -> str:
        """[see superclass]"""
        return "text"

    def start_drawing(self, title: str, **kwargs) -> None:
        """
        Write the title to :attr:`.out`.

        :param title: the title of the drawing
        """
        print(f"{f' {title} ':*^{self.width}s}", file=self.out)


__tracker.validate()



"""
Utilities for rendering text.
"""
import logging
from typing import Any, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd

from pytools.api import AllTracker

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = ["CharacterMatrix", "format_table"]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Type definitions
#

_TextCoordinates = Tuple[Union[int, slice], Union[int, slice]]


#
# Classes
#


class CharacterMatrix:
    """
    A matrix of characters, indexed by rows and columns.

    The matrix is initialised with space characters (`" "`).

    Characters can be "painted" in the matrix using 2D index expressions:

    - ``matrix[r, c] = chr`` assigns character `chr` at position `(r, c)`
    - ``matrix[r, c1:c2] = str`` writes string `str` at positions
      `(r, c1) … (r, c2 – 1)`; excess characters in `str` are clipped if ``len(s)`` is
      greater than `c2 – c1`
    - ``matrix[r, c1:c2] = chr`` (where `chr` is a single character) repeats `chr` at
      every position `(r, c1) … (r, c2 – 1)`
    - ``matrix[r1:r2, …] = …`` applies the same insertion at each of rows `r1 … r2 – 1`
    - full slice notation is supported so even slices of shape ``start:stop:step`` work
      as expected

    :param n_rows: the matrix height
    :param n_columns: the matrix width
    """

    def __init__(self, n_rows: int, n_columns: int):
        if n_columns <= 0:
            raise ValueError(f"arg width must be positive but is {n_columns}")
        if n_rows <= 0:
            raise ValueError(f"arg height must be positive but is {n_rows}")
        self._n_columns = n_columns
        self._matrix = [[" " for _ in range(n_columns)] for _ in range(n_rows)]

    @property
    def n_rows(self) -> int:
        """
        The height of this matrix.

        Same as ``len(self)``.
        """
        return len(self._matrix)

    @property
    def n_columns(self) -> int:
        """
        The width of this matrix.
        """
        return self._n_columns

    def lines(self) -> Iterable[str]:
        """
        Get this character matrix as strings representing the matrix rows.

        :return: the rows in this matrix as strings
        """
        return ("".join(line) for line in self._matrix)

    @staticmethod
    def __key_as_slices(key: _TextCoordinates) -> Tuple[slice, slice]:
        def _to_slice(index: Union[int, slice]) -> slice:
            if isinstance(index, int):
                return slice(index, index + 1)
            else:
                return index

        if not isinstance(key, tuple) or len(key) != 2:
            raise ValueError(f"expected (row, column) tuple but got {key}")

        row, column = key
        return _to_slice(row), _to_slice(column)

    def __str__(self) -> str:
        return "\n".join(self.lines())

    def __len__(self) -> int:
        return self.n_rows

    def __getitem__(self, key: _TextCoordinates):
        rows, columns = self.__key_as_slices(key)
        return "\n".join("".join(line[columns]) for line in self._matrix[rows])

    def __setitem__(self, key: _TextCoordinates, value: Any) -> None:
        rows, columns = self.__key_as_slices(key)
        value = str(value)
        single_char = len(value) == 1
        positions = range(*columns.indices(self.n_columns))
        for line in self._matrix[rows]:
            if single_char:
                for pos in positions:
                    line[pos] = value
            else:
                for pos, char in zip(positions, value):
                    line[pos] = char


_ALIGNMENT_OPTIONS = ["<", "^", ">"]


def format_table(
    headings: Sequence[str],
    data: Union[pd.DataFrame, np.ndarray, Sequence[Sequence[Any]]],
    formats: Optional[Sequence[Optional[str]]] = None,
    alignment: Optional[Sequence[Optional[str]]] = None,
) -> str:
    """
    Print a formatted text table.

    :param headings: the table headings
    :param data: the table data, as an array-like with shape `[n_rows, n_columns]`
    :param formats: formatting strings for data in each row (optional);
        uses ``str()`` conversion for any formatting strings stated as ``None``
    :param alignment: text alignment for each column (optional); use ``"<"`` to align
        left, ``"="`` to center, ``">"`` to align right (defaults to left alignment)
    :return: the formatted table as a multi-line string
    """
    n_columns = len(headings)

    if formats is None:
        formats = [None] * n_columns
    elif len(formats) != n_columns:
        raise ValueError("arg formats must have the same length as arg headings")

    if alignment is None:
        alignment = ["<"] * n_columns
    elif len(alignment) != n_columns:
        raise ValueError("arg alignment must have the same length as arg headings")
    elif not all(align in _ALIGNMENT_OPTIONS for align in alignment):
        raise ValueError(
            f"arg alignment must only contain alignment options "
            f'{", ".join(_ALIGNMENT_OPTIONS)}'
        )

    def _formatted(item: Any, format_string: str) -> str:
        if format_string is None:
            return str(item)
        else:
            return f"{item:{format_string}}"

    def _iterate_row_data() -> Iterable[Sequence]:
        if isinstance(data, pd.DataFrame):
            return (row for _, row in data.iterrows())
        else:
            return iter(data)

    def _make_row(items: Sequence):
        if len(items) != n_columns:
            raise ValueError(
                "rows in data matrix must have the same length as arg headings"
            )
        return [
            _formatted(item, format_string)
            for item, format_string in zip(items, formats)
        ]

    body_rows = [_make_row(items) for items in _iterate_row_data()]

    column_widths = [
        max(column_lengths)
        for column_lengths in zip(
            *(
                (len(item) for item in row)
                for row in (
                    headings,
                    *[_make_row(items) for items in _iterate_row_data()],
                )
            )
        )
    ]

    dividers = ["=" * column_width for column_width in column_widths]

    def _format_rows(rows: List[List[str]], align: bool):
        return (
            "  ".join(
                (
                    f'{item:{align_char if align else ""}{column_width}s}'
                    for item, align_char, column_width in zip(
                        row, alignment, column_widths
                    )
                )
            )
            for row in rows
        )

    return "\n".join(
        (
            *(_format_rows(rows=[headings, dividers], align=False)),
            *(_format_rows(rows=body_rows, align=True)),
            "",
        )
    )


__tracker.validate()
